import matplotlib.pyplot as _plt
import numpy as _np

import pybdsim as _pybdsim
import pybdsim.Field._Field

def _ArrowSize(d):
    """
    :param d: pybdsim.Field.Field instance
    :type  d: pybdsim.Field.Field
    """
    nDim = d.nDimensions
    h = d.header
    result = _np.inf
    for i in range(nDim):
        key = d.columns[i].lower()
        step = (h[key+'max'] - h[key+'min']) / h['n'+key]
        result = _np.min([result, step])
    return result

class FourDData(object):
    """
    Class purely to simplify plotting of fields. Not for general use.
    """
    def __init__(self, filename, xind=0, yind=1, zind=2, tind=3):
        if type(filename) is str:
            d = _pybdsim.Field.Load(filename)
        elif isinstance(filename, _pybdsim.Field._Field.Field):
            d = filename.data
        else:
            d = filename
            
        # '...' fills in unknown number of dimensions with ':' meaning
        # all of that dimension
        if (xind >= 0):
            self.x  = d[..., xind].flatten()
        if (yind >= 0):
            self.y  = d[..., yind].flatten()
        if (zind >= 0):
            self.z  = d[..., zind].flatten()
        if (tind >= 0):
            self.t  = d[..., tind].flatten()

        # index from end as we don't know the dimensionality
        self.fx = d[...,-3].flatten()
        self.fy = d[...,-2].flatten()
        self.fz = d[...,-1].flatten()

        self.mag = _np.sqrt(self.fx**2 + self.fy**2 + self.fz**2)

class ThreeDData(FourDData):
    """
    Class purely to simplify plotting of fields. Not for general use.
    """
    def __init__(self, filename):
        FourDData.__init__(self, filename, tind=-1)

class TwoDData(FourDData):
    """
    Class purely to simplify plotting of fields. Not for general use.
    """
    def __init__(self, filename):
        FourDData.__init__(self, filename, tind=-1, zind=-1)

class OneDData(FourDData):
    """
    Class purely to simplify plotting of fields. Not for general use.
    """
    def __init__(self, filename):
        FourDData.__init__(self, filename, tind=-1, zind=-1, yind=-1)

def _Niceties(xlabel, ylabel, zlabel="", flipX=False, aspect='equal'):
    if flipX:
        cx = _plt.xlim()
        _plt.xlim(cx[1],cx[0]) # plot backwards in effect
    _plt.xlabel(xlabel)
    _plt.ylabel(ylabel)
    _plt.colorbar(label=zlabel)
    ax = _plt.gca()
    ax.set_aspect(aspect)
    _plt.tight_layout()


def Plot1DFxFyFz(filename):
    """
    Plot a bdsim 1D field map file.

    :param filename: name of the field map file or object
    :type filename: str, pybdsim.Field._Field.Field1D instance
    """
    if type(filename) is str:
        d = _pybdsim.Field.Load(filename)
    elif isinstance(filename, _pybdsim.Field._Field.Field1D):
        d = filename
    else:
        raise TypeError("'filename' must be either str or Field1D")

    f = _plt.figure(figsize=(7.5,4))
    axFz = f.add_subplot(313)
    axFx = f.add_subplot(311, sharex=axFz)
    axFy = f.add_subplot(312, sharex=axFz)

    axFx.plot(d.data[:,0], d.data[:,1],'b')
    axFx.plot(d.data[:,0], d.data[:,1],'b.')
    axFy.plot(d.data[:,0], d.data[:,2],'g')
    axFy.plot(d.data[:,0], d.data[:,2],'g.')
    axFz.plot(d.data[:,0], d.data[:,3],'r')
    axFz.plot(d.data[:,0], d.data[:,3],'r.')

    axFx.set_ylabel('B$_x$ (T)')
    axFy.set_ylabel('B$_y$ (T)')
    axFz.set_ylabel('B$_z$ (T)')
    axFz.set_xlabel(d.columns[0]+' (cm)')

    _plt.setp(axFx.get_xticklabels(), visible=False)
    _plt.setp(axFy.get_xticklabels(), visible=False)
    _plt.tight_layout()

def Plot2DXY(filename, scale=None, title=None, flipX=False, firstDimension="X", secondDimension="Y", aspect='equal', figsize=(6,5)):
    """
    Plot a bdsim field map file using the X,Y plane.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param scale: numerical scaling for quiver plot arrow lengths.
    :type scale: float
    :param title: title for plot
    :type title: str
    :param flipX: whether to plot x backwards to match the right hand coordinate system of Geant4.
    :type flipX: bool
    :param firstDimension: Label of first dimension, e.g. "X"
    :type firstDimension: str
    :param secondDimension: Label of second dimension, e.g. "Z"
    :type secondDimension: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    """
    d = TwoDData(filename)
    _plt.figure(figsize=figsize)
    _plt.quiver(d.x,d.y,d.fx,d.fy,d.mag,cmap=_plt.cm.magma,pivot='mid',scale=scale)
    if title:
        _plt.title(title)
    _Niceties(firstDimension+' (cm)', secondDimension+' (cm)', zlabel="|$B_{x,y}$| (T)", flipX=flipX, aspect=aspect)


def Plot2DXYMagnitude(filename, title=None, flipX=False, firstDimension="X", secondDimension="Y", aspect="equal", zlabel="|$B_{x,y}$| (T)", figsize=(6,5)):
    """
    Plot a the magnitude of a 2D bdsim field map file using any two planes.

    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param title: title for plot
    :type title: str, None
    :param flipX: whether to plot x backwards to match the right hand coordinate system of Geant4.
    :type flipX: bool
    :param firstDimension: Name of first dimension, e.g. "X"
    :type firstDimension: str
    :param secondDimension: Name of second dimension, e.g. "Z"
    :type secondDimension: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    :param zlabel: Label for colour bar
    :type zlabel: str
    """
    flip = False
    if type(filename) is str:
        doriginal = _pybdsim.Field.Load(filename)
        d = doriginal.data
        flip = doriginal.flip
    elif isinstance(filename, _pybdsim.Field._Field.Field):
        doriginal = filename
        d = filename.data
        flip = doriginal.flip
    else:
        raise ValueError("Invalid type of data")
    
    _plt.figure(figsize=figsize)

    # assumes the columns are X Y Fx Fy Fz
    bmag = _np.sqrt(d[:,:,2]**2 + d[:,:,3]**2)

    ext = [_np.min(d[:,:,0]), _np.max(d[:,:,0]), _np.min(d[:,:,1]), _np.max(d[:,:,1])]

    # the data will write out flipped but we need to draw it the right way
    theData = bmag.T if flip else bmag
    _plt.imshow(theData, extent=ext, origin='lower', aspect=aspect, interpolation='none')

    if title:
        _plt.title(title)

    fd = firstDimension
    sd = secondDimension
    _Niceties(fd+' (cm)', sd+' (cm)', zlabel=zlabel, flipX=flipX, aspect=aspect)


def Plot2D(filename, scale=None, title=None, flipX=False, flipY=False, firstDimension="X", secondDimension="Y", aspect="equal"):
    """
    Plot a bdsim field map file using any two planes. The corresponding
    field components are plotted (e.g. X:Z -> Fx:Fz).

    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param scale: numerical scaling for quiver plot arrow lengths.
    :type scale: float
    :param title: title for plot
    :type title: str
    :param flipX: whether to plot x backwards to match the right hand coordinate system of Geant4.
    :type flipX: bool
    :param firstDimension: Name of first dimension, e.g. "X"
    :type firstDimension: str
    :param secondDimension: Name of second dimension, e.g. "Z"
    :type secondDimension: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    """
    if type(filename) is str:
        doriginal = _pybdsim.Field.Load(filename)
        d = doriginal.data
    elif isinstance(filename, _pybdsim.Field._Field.Field):
        doriginal = filename
        d = filename.data
    else:
        raise ValueError("Invalid type of data")
    _plt.figure()
    assert(firstDimension != secondDimension)
    iInd = ['x', 'y', 'z', 't'].index(firstDimension.lower())
    jInd = ['x', 'y', 'z', 't'].index(secondDimension.lower())
    ci = d[:, :, 0].flatten()
    cj = d[:, :, 1].flatten()
    fi = d[:, :, iInd+2].flatten()
    fj = d[:, :, jInd+2].flatten()
    if scale is None:
        scale = _ArrowSize(doriginal)
    fmag = _np.hypot(fi,fj)
    fi /= fmag
    fj /= fmag
    _plt.quiver(ci, cj, fi, fj, fmag, cmap=_plt.cm.magma, pivot='mid', scale=1.0/scale, units='xy', scale_units='xy')
    if title:
        _plt.title(title)
    fd = firstDimension
    sd = secondDimension
    _Niceties(fd + ' (cm)', sd + ' (cm)', zlabel="|$B_{"+fd+","+sd+"}$| (T)", flipX=flipX, aspect=aspect)

def Plot2DXYStream(filename, density=1, zIndexIf3D=0, useColour=True, aspect='equal'):
    """
    Plot a bdsim field map file using the X,Y plane as a stream plot and plotting Fx, Fy.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D or Field3D instance
    :param density: arrow density (default=1) for matplotlib streamplot
    :type density: float
    :param zIndexIf3D: index in Z if using 3D field map (default=0)
    :type zIndexIf3D: int
    :param useColour: use magnitude of field as colour.
    :type useColour: bool\
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str

    Note, matplotlibs streamplot may raise an exception if the field is entriely 0 valued.
    """
    if type(filename) is str:
        d = _pybdsim.Field.Load(filename)
    elif isinstance(filename, _pybdsim.Field._Field.Field):
        d = filename.data
    else:
        raise ValueError("Invalid type of data")

    shape = _np.shape(d.data)
    zInd = zIndexIf3D
    if len(shape) == 3:
        cx = d[0,:,0]
        cy = d[:,0,1]
        fx = d[:,:,2]
        fy = d[:,:,3]
    elif len(shape) == 4:
        cx = d[zInd,0,:,0]
        cy = d[zInd,:,0,1]
        fx = d[zInd,:,:,3]
        fy = d[zInd,:,:,4]
    else:
        raise ValueError("Currently only 2D and 3D field maps supported.")

    # modern matplotlib's streamplot has a very strict check on the spacing
    # of points being equal, which they're meant. However, it is too strict
    # given the precision of incoming data, So, knowing here they're linearly
    # spaced, we regenerate again for the purpose of this plot.
    cx = _np.linspace(_np.min(cx), _np.max(cx), len(cx))
    cy = _np.linspace(_np.min(cy), _np.max(cy), len(cy))
    
    _plt.figure()
    if useColour:
        mag = _np.sqrt(fx**2 + fy**2)
        _plt.streamplot(cx,cy,fx,fy,color=mag,cmap=_plt.cm.magma,density=density)
    else:
        _plt.streamplot(cx,cy,fx,fy,density=density)
    _Niceties('X (cm)', 'Y (cm)', zlabel="|$B_{x,y}$| (T)", aspect=aspect)

def Plot2DXZStream(filename, density=1, yIndexIf3D=0, useColour=True, aspect='equal'):
    """
    Plot a bdsim field map file using the X,Z plane as a stream plot and plotting Fx, Fz.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D or Field3D instance
    :param density: arrow density (default=1) for matplotlib streamplot
    :type density: float
    :param yIndexIf3D: index in Z if using 3D field map (default=0)
    :type yIndexIf3D: int
    :param useColour: use magnitude of field as colour.
    :type useColour: bool
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str

    Note, matplotlibs streamplot may raise an exception if the field is entriely 0 valued.
    """
    if type(filename) is str:
        d = _pybdsim.Field.Load(filename)
    elif isinstance(filename, _pybdsim.Field._Field.Field):
        d = filename.data
    else:
        raise ValueError("Invalid type of data")

    shape = _np.shape(d.data)
    yInd = yIndexIf3D
    if len(shape) == 3:
        cx = d[0,:,0]
        cz = d[:,0,1] # still 2d data
        fx = d[:,:,2]
        fz = d[:,:,4]
    elif len(shape) == 4:
        cx = d[yInd,0,:,0]
        cz = d[yInd,:,0,2]
        fx = d[yInd,:,:,3]
        fz = d[yInd,:,:,5]
    else:
        raise ValueError("Currently only 2D and 3D field maps supported.")

    # modern matplotlib's streamplot has a very strict check on the spacing
    # of points being equal, which they're meant. However, it is too strict
    # given the precision of incoming data, So, knowing here they're linearly
    # spaced, we regenerate again for the purpose of this plot.
    cx = _np.linspace(_np.min(cx), _np.max(cx), len(cx))
    cz = _np.linspace(_np.min(cz), _np.max(cz), len(cz))
    
    _plt.figure()
    if useColour:
        mag = _np.sqrt(fx**2 + fz**2)
        _plt.streamplot(cx,cz,fx,fz,color=mag,cmap=_plt.cm.magma,density=density)
    else:
        _plt.streamplot(cx,cz,fx,fz,density=density)
    _Niceties('X (cm)', 'Z (cm)', zlabel="|$B_{x,z}$| (T)", aspect=aspect)

def Plot2DXYConnectionOrder(filename):
    """
    Plot a point in orange and a line in blue (default matplotlib colours)
    for each location in the field map. If the field map is constructed
    correctly, this should show a set of lines with diagonals between them.
    The other plots with the arrows are independent of order unlike when
    BDSIM loads the fields. So you might see an OK field map, but it could
    be wrong if handwritten.
    """
    d = TwoDData(filename)
    _plt.figure()
    _plt.plot(d.x,d.y)
    _plt.plot(d.x,d.y,'.')
    _plt.xlabel('X (cm)')
    _plt.ylabel('Y (cm)')
    _plt.tight_layout()

def Plot2DXYComponent(filename, componentIndex=2, scale=None, title=None, flipX=False, aspect='equal'):
    """
    Plot a bdsim field map file use the X,Y plane, but plotting By component.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param componentIndex: index of field component (0,1,2) for Fx, Fy, Fz
    :type componentIndex: int
    :param scale: numerical scaling for quiver plot arrow lengths.
    :type scale: float
    :param title: title for plot
    :type title: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    """
    if type(filename) is str:
        d = _pybdsim.Field.Load(filename)
    elif isinstance(filename, _pybdsim.Field._Field.Field2D):
        d = filename
    else:
        raise TypeError("'filename' must be either str or Field2D")

    acceptableIndices = {0,1,2}
    assert(componentIndex in acceptableIndices)

    label = ["x", "y", "z"][componentIndex]

    ci = -3 + componentIndex
    
    _plt.figure()
    ext = [_np.min(d.data[:,:,0]),_np.max(d.data[:,:,0]),_np.min(d.data[:,:,1]),_np.max(d.data[:,:,1])]
    _plt.imshow(d.data[:,:,ci], extent=ext, origin='lower', aspect='equal', interpolation='none', cmap=_plt.cm.magma)
    if title:
        _plt.title(title)
    _Niceties(d.columns[0]+' (cm)', d.columns[1]+' (cm)', zlabel="$B_"+label+"$ (T)", flipX=flipX, aspect=aspect)

def Plot2DXYBx(filename, scale=None, title=None, flipX=False, aspect='equal'):
    """
    Plot a bdsim field map file use the X,Y plane, but plotting By component.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param scale: numerical scaling for quiver plot arrow lengths.
    :type scale: float
    :param title: title for plot
    :type title: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    """
    Plot2DXYComponent(filename, 0, scale, title, flipX, aspect)

def Plot2DXYBy(filename, scale=None, title=None, flipX=False, aspect='equal'):
    """
    Plot a bdsim field map file use the X,Y plane, but plotting By component.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param scale: numerical scaling for quiver plot arrow lengths.
    :type scale: float
    :param title: title for plot
    :type title: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    """
    Plot2DXYComponent(filename, 1, scale, title, flipX, aspect)

def Plot2DXYBz(filename, scale=None, title=None, flipX=False, aspect='equal'):
    """
    Plot a bdsim field map file use the X,Y plane, but plotting By component.
    
    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field2D instance
    :param scale: numerical scaling for quiver plot arrow lengths.
    :type scale: float
    :param title: title for plot
    :type title: str
    :param aspect: Matplotlib axes aspect (e.g. 'auto' or 'equal')
    :type aspect: str
    """
    Plot2DXYComponent(filename, 2, scale, title, flipX, aspect)


def Plot2DXYFxFyFz(filename, title=None, aspect="auto", extent=None, **imshowKwargs):
    """
    Plot Fx,Fy,Fz components of a field separately as a function of X,Y.

    :param filename: name of field map file or object
    :type filename: str, pybdsim.Field._Field.Field1D instance
    :param title: optional title for plot
    :type title: None, str
    :param aspect: aspect ratio for matplotlib imshow
    :type aspect: str
    :param extent: list or tuple of (xmin,xmax,ymin,ymax) for each plot (optional)
    :type extent: list,tuple
    """
    imshowKwargs['aspect'] = aspect
    if type(filename) is str:
        fd = _pybdsim.Field.Load(filename)
    elif isinstance(filename, _pybdsim.Field._Field.Field2D):
        fd = filename
    else:
        raise TypeError("'filename' must be either str or Field2D")
    a = fd.data
    
    f   = _plt.figure(figsize=(7.5,4))
    ax  = f.add_subplot(131)
    ax2 = f.add_subplot(132)
    ax3 = f.add_subplot(133)

    xmin = _np.min(a[:,:,0])
    xmax = _np.max(a[:,:,0])
    ymin = _np.min(a[:,:,1])
    ymax = _np.max(a[:,:,1])
    if extent is None:
        extent = [_np.min(a[:,:,0]), _np.max(a[:,:,0]), _np.min(a[:,:,1]), _np.max(a[:,:,1])]
    imshowKwargs['extent'] = extent

    # determine a consistent colour min and max value for all three subplots
    if 'vmin' not in imshowKwargs:
        imshowKwargs['vmin'] = _np.min(a[:,:,3:])
    if 'vmax' not in imshowKwargs:
        imshowKwargs['vmax'] = _np.max(a[:,:,3:])
    
    ax.imshow(a[:,:,2], interpolation='None', origin='lower', **imshowKwargs)
    ax.set_xlabel('X (cm)')
    ax.set_ylabel('Y (cm)')
    ax.set_title('X-Component',size='medium')
    
    im = ax2.imshow(a[:,:,3], interpolation='None', origin='lower', **imshowKwargs)
    ax2.set_xlabel('X (cm)')
    ax2.set_title('Y-Component',size='medium')
    ax2.get_yaxis().set_ticks([])

    im = ax3.imshow(a[:,:,4], interpolation='None', origin='lower', **imshowKwargs)
    ax3.set_xlabel('X (cm)')
    ax3.set_title('Z-Component',size='medium')
    ax3.get_yaxis().set_ticks([])

    f.subplots_adjust(left=0.09, right=0.85, top=0.86, bottom=0.12, wspace=0.02)
    cbar_ax = f.add_axes([0.88, 0.15, 0.05, 0.7])
    f.colorbar(im, cax=cbar_ax)

    if title:
        _plt.suptitle(title, size='x-large')

def Plot3DXY(filename, scale=None):
    """
    Plots (B_x,B_y) as a function of x and y.
    """
    d = ThreeDData(filename)
    _plt.figure()
    _plt.quiver(d.x,d.y,d.fx,d.fy,d.mag,cmap=_plt.cm.magma,pivot='mid',scale=scale)
    _Niceties('X (cm)', 'Y (cm)')

def Plot3DXZ(filename, scale=None):
    """
    Plots (B_x,B_z) as a function of x and z.
    """
    d = ThreeDData(filename)
    _plt.figure()
    _plt.quiver(d.x,d.z,d.fx,d.fz,d.mag,cmap=_plt.cm.magma,pivot='mid',scale=scale)
    _Niceties('X (cm)', 'Z (cm)')



"""
See Options class inside this module.

"""

def ProtonColliderOptions():
    a = Options()
    a.SetPhysicsList('FTFP_BERT')
    a.SetBeamPipeThickness(5,'mm')
    a.SetOuterDiameter(0.5, 'm')
    a.SetTunnelRadius(2, 'm')
    a.SetNGenerate(100)
    a.SetBeamPipeRadius(5,'cm')
    a.SetBuildTunnel(True)
    a.SetBuildTunnelFloor(True)
    a.SetTunnelRadius(2,'m')
    a.SetTunnelThickness(50,'cm')
    a.SetTunnelOffsetX(1,'m')
    a.SetTunnelFloorOffset(1,'m')
    return a

def ElectronColliderOptions():
    a = Options()
    a.SetPhysicsList('em')
    a.SetBeamPipeThickness(5,'mm')
    a.SetOuterDiameter(1,'m')
    a.SetTunnelRadius(2,'m')
    a.SetNGenerate(100)
    a.SetDefaultRangeCut(0.25,"m")
    a.SetBeamPipeRadius(5,'cm')
    a.SetBuildTunnel(True)
    a.SetBuildTunnelFloor(True)
    a.SetTunnelRadius(2,'m')
    a.SetTunnelThickness(50,'cm')
    a.SetTunnelOffsetX(1,'m')
    a.SetTunnelFloorOffset(1,'m')
    return a

class Options(dict):
    """
    Inherits a dict. Converting to a string or using
    ReturnOptionsString() will give a suitable GMAD
    string to write out to a file.

    o = pybdsim.Options.Options()
    o["trajectoryConnect"] = 1
    o["aper1"] = (5, 'm')
    str(o)
    'option,\ttrajectoryConnect=1,\n\taper1=5*m;'

    There is no checking on the option if using []. A tuple
    of (value, unitsString) can be used too resulting in
    value*unitsString.
    """
    def __init__(self,*args,**kwargs):
        dict.__init__(self,*args,**kwargs)

    def __repr__(self):
        return self.ReturnOptionsString()

    def SetGeneralOption(self, option, value):
        self[option] = value

    def ReturnOptionsString(self):
        s = ''
        if len(list(self.keys())) == 0:
            #print 'No options set - empty string'
            return s
        
        numOptions=0
        for k,v in self.items():
            if type(v) is tuple:
                vs = str(v[0]) + "*" + v[1]
            else:
                vs = str(v)
            s += ', \n\t'+str(k)+'='+vs
            numOptions += 1
        s += ';'
        s2 = s.split('\n')
        s3 = 'option,\t'+s2[1].replace('\t','').replace('\n','').replace(',','').strip()
        if numOptions == 1:
            s3 += '\n'
        else:
            s3 += ',\n'
        s4 = '\n'.join(s2[2:])
        st = s3+s4
        return st

    def SetNGenerate(self, nparticles=1):
        self['ngenerate'] = nparticles

    def SetPhysicsList(self, physicslist=''):
        physicslistlist = [
            'all_particles',
            'annihi_to_mumu',
            'charge_exchange',
            'cherenkov',
            'cuts_and_limits',
            'decay',
            'decay_radioactive',
            'em',
            'em_extra',
            'em_livermore',
            'em_livermore_polarised',
            'em_low_ep',
            'em_penelope',
            'em_low', # alias to em_penelope
            'em_ss',
            'em_wvi',
            'em_1',
            'em_2',
            'em_3',
            'em_4',
            'ftfp_bert',
            'ftfp_bert_hp',
            'hadronic', # alias to ftfp_bert
            'hadronichp', # alias to ftfp_bert_hp
            'hadronic_elastic',
            'hadronic_elastic_d',
            'hadronic_elastic_h',
            'hadronic_elastic_hp',
            'hadronic_elastic_lend',
            'hadronic_elastic_xs',
            'ion',
            'ion_binary',
            'ion_elastic',
            'ion_elastic_qmd',
            'ion_em_dissociation',
            'ion_inclxx',
            'lw',
            'muon',
            'neutron_tracking_cut',
            'optical',
            'qgsp_bert',
            'qgsp_bert_hp',
            'qgsp_bic',
            'qgsp_bic_hp',
            'shielding',
            'stopping',
            'synch_rad',
            'synchrad', # alias to synch_rad
            'em_gs',
            'decay_spin',
            'ion_php',
            'decay_muonic_atom',
            'channelling',
            'dna',
            'dna_1',
            'dna_2',
            'dna_2',
            'dna_3',
            'dna_4',
            'dna_5',
            'dna_6',
            'dna_7',
            'radioactivation',
            'shielding_lend'
            ]
        if len(physicslist.split()) == 1:
            if physicslist not in physicslistlist and not physicslist.startswith("g4"):
                print('Warning: unknown physicslist: '+physicslist)
            self['physicsList'] = '"' + str(physicslist) + '"'
        else:
            splitphysicslist = physicslist.split()
            for token in splitphysicslist :
                if token not in physicslistlist :
                    print('Warning: unknown physicslist: '+physicslist)

            self['physicsList'] = '"' + str(physicslist) + '"'

    def SetBeamPipeRadius(self, beampiperadius=5.0, unitsstring='cm'):
        self['beampipeRadius'] = str(beampiperadius) + '*' +unitsstring
        
    def SetOuterDiameter(self, outerdiameter=2.0, unitsstring='m'):
        self['outerDiameter'] = str(outerdiameter) + '*' + unitsstring

    def SetTunnelRadius(self, tunnelradius=2.0, unitsstring='m'):
        self['tunnelRadius'] = str(tunnelradius) + '*' + unitsstring

    def SetBeamPipeThickness(self, bpt, unitsstring='mm'):
        self['beampipeThickness'] = str(bpt) + '*' + unitsstring

    def SetPipeMaterial(self, bpm):
        self['pipeMaterial'] = '"' + str(bpm) + '"'

    def SetVacuumMaterial(self, vm):
        self['vacMaterial'] = '"' + str(vm) + '"'

    def SetVacuumPressure(self, vp):
        """
        Vacuum pressure in bar
        """
        self['vacuumPressure'] = str(vp)

    def SetBuildTunnel(self, tunnel=False):
        self['buildTunnel'] = int(tunnel)

    def SetBuildTunnelFloor(self, tunnelfloor=False):
        self['buildTunnelFloor'] = int(tunnelfloor)

    def SetTunnelThickness(self, tt=1.0, unitsstring='m'):
        self['tunnelThickness'] = str(tt) + '*' + unitsstring

    def SetSoilThickness(self, st=4.0, unitsstring='m'):
        self['tunnelSoilThickness'] = str(st) + '*' + unitsstring

    def SetTunnelMaterial(self, tm):
        self['tunnelMaterial'] = '"' + str(tm) + '"'

    def SetSoilMaterial(self, sm):
        self['soilMaterial'] = '"' + str(sm) + '"'

    def SetTunnelOffsetX(self, offset=0.0, unitsstring='m'):
        self['tunnelOffsetX'] = str(offset) + '*' + unitsstring

    def SetTunnelOffsetY(self, offset=0.0, unitsstring='m'):
        self['tunnelOffsetY'] = str(offset) + '*' + unitsstring

    def SetTunnelFloorOffset(self, offset=1.0, unitsstring='m'):
        self['tunnelFloorOffset'] = str(offset) + '*' + unitsstring

    def SetSamplerDiameter(self, radius=10.0, unitsstring='m'):
        self['samplerDiameter'] = str(radius) + '*' + unitsstring

    def SetBLMRadius(self, radius=5.0, unitsstring='cm'):
        self['blmRad'] = str(radius) + '*' + unitsstring

    def SetBLMLength(self, length=50.0, unitsstring='cm'):
        self['blmLength'] = str(length) + '*' + unitsstring

    def SetIncludeIronMagField(self, iron=True):
        self['includeIronMagFields'] = int(iron)

    def SetDontSplitSBends(self, dontsplitsbends=False):
        self['dontSplitSBends'] = int(dontsplitsbends)

    def SetDeltaChord(self, dc=0.001, unitsstring='m'):
        self['deltaChord'] = str(dc) + '*' + unitsstring

    def SetDeltaIntersection(self, di=10.0, unitsstring='nm'):
        self['deltaIntersection'] = str(di) + '*' + unitsstring

    def SetChordStepMinimum(self, csm=1.0, unitsstring='nm'):
        self['chordStepMinimum'] = str(csm) + '*' + unitsstring

    def SetLengthSafety(self, ls=10.0, unitsstring='um'):
        self['lengthSafety'] = str(ls) + '*' + unitsstring

    def SetMinimumEpsilonStep(self, mes=10.0, unitsstring='nm'):
        self['minimumEpsilonStep'] = str(mes) + '*' + unitsstring

    def SetMaximumEpsilonStep(self, mes=1.0, unitsstring='m'):
        self['maximumEpsilonStep'] = str(mes) + '*' + unitsstring

    def SetDeltaOneStep(self, dos=10.0, unitsstring='nm'):
        self['deltaOneStep'] = str(dos) + '*' + unitsstring

    def SetMaximumStepLength(self, msl=20.0, unitsstring='m'):
        self['maximumStepLength'] = str(msl) + '*' + unitsstring

    def SetMaximumTrackingTime(self, mtt=-1.0, unitsstring='s'):
        self['maximumTrackingTime'] = str(mtt) + '*' + unitsstring

    def SetIntegratorSet(self, integratorSet='"bdsim"'):
        self['integratorSet'] = integratorSet

    def SetThresholdCutCharged(self, tcc=100.0, unitsstring='MeV'):
        self['thresholdCutCharged'] = str(tcc) + '*' + unitsstring

    def SetThresholdCutPhotons(self, tcp=1.0, unitsstring='MeV'):
        self['thresholdCutPhotons'] = str(tcp) + '*' + unitsstring

    def SetStopSecondaries(self, stop=True):
        self['stopSecondaries'] = int(stop)

    def SetSynchRadiationOn(self, on=True):
        self['synchRadOn'] = int(on)

    def SetTrackSRPhotons(self, track=True):
        self['srTrackPhotons'] = int(track)

    def SetSRLowX(self, lowx=True):
        self['srLowX'] = int(lowx)

    def SetSRMultiplicity(self, srm=2.0):
        self['srMultiplicity'] = srm

    def SetProductionCutPhotons(self, pc=100.0, unitsstring='keV'):
        self['prodCutPhotons'] = str(pc) + '*' + unitsstring

    def SetProductionCutElectrons(self, pc=100.0, unitsstring='keV'):
        self['prodCutElectrons'] = str(pc) + '*' + unitsstring

    def SetProductionCutPositrons(self, pc=100.0, unitsstring='keV'):
        self['prodCutPositrons'] = str(pc) + '*' + unitsstring

    def SetCherenkovOn(self, on=True):
        self['turnOnCerenkov'] = int(on)

    def SetDefaultRangeCut(self, drc=0.7, unitsstring='mm'):
        self['defaultRangeCut'] = str(drc) + '*' + unitsstring

    def SetGamma2MuonEnahncementFactor(self, ef=1.0):
        self['gammToMuFe'] = ef

    def SetEPAnnihilation2MuonEnhancementFactor(self, ef=1.0):
        self['annihiToMuFe'] = ef

    def SetEPAnnihilation2HadronEnhancementFactor(self, ef=1.0):
        self['eetoHadronsFe'] = ef

    def SetEMLeadParticleBiasing(self, on=True):
        self['useEMLPB'] = int(on)

    def SetLPBFraction(self, fraction=0.5):
        self['LPBFraction'] = fraction

    def SetRandomSeed(self, rs=0):
        self['randomSeed'] = rs

    def SetNGenerate(self, nparticles=10):
        self['ngenerate'] = nparticles

    def SetWritePrimaries(self, on=True):
        self['writePrimaries'] = int(on)

    def SetELossHistBinWidth(self,width):
        self['elossHistoBinWidth'] = width

    def SetSensitiveBeamlineComponents(self, on=True):
        self['sensitiveBeamLineComponents'] = int(on)

    def SetSensitiveBeamPipe(self, on=True):
        self['sensitiveBeamPipe'] = int(on)

    def SetSenssitiveBLMs(self, on=True):
        self['sensitiveBLMs'] = int(on)

    def SetStoreTrajectory(self, on=True):
        self['storeTrajectory'] = int(on)

    def SetStoreTrajectoryParticle(self, particle="muon"):
        self['storeTrajectoryParticle'] = particle

    def SetMagnetGeometryType(self, magnetGeometryType='"none"'):
        self['magnetGeometryType'] = magnetGeometryType

    def SetTrajectoryCutGTZ(self, gtz=0.0, unitsstring='m'):
        self['trajCutGTZ'] = str(gtz) + '*' + unitsstring

    def SetTrajectoryCutLTR(self, ltr=10.0, unitsstring='m'):
        self['trajCutLTR'] = str(ltr) + '*' + unitsstring

    def SetPrintModuloFraction(self, pmf=1e-2):
        self['printModuloFraction'] = pmf

    def SetNPerFile(self, nperfile=100):
        self['nperfile'] = nperfile

    def SetNLinesIgnore(self, nlines=0):
        self['nlinesIgnore'] = nlines

    def SetIncludeFringeFields(self, on=True):
        self['includeFringeFields'] = int(on)

    def SetDefaultBiasVaccum(self, biases=""):
        self["defaultBiasVacuum"] = biases

    def SetDefaultBiasMaterial(self, biases=""):
        self["defaultBiasMaterial"] = biases
        
    def SetBeamlineS(self,beamlineS=0, unitsstring='m'):
        self["beamlineS"] = str(beamlineS) + '*' +unitsstring



from copy import deepcopy as _deepcopy
import numpy as _np
import pymadx as _pymadx
import warnings as _warnings
from .. import Builder as _Builder
from ..Options import Options as _Options
from .. import Beam as _Beam
from .. import Data as _Data
import pybdsim._General

_requiredKeys = frozenset([
    'L', 'ANGLE', 'KSI',
    'K1L', 'K2L', 'K3L', 'K4L', 'K5L', 'K6L',
    'K1SL', 'K2SL', 'K3SL', 'K4SL', 'K5SL', 'K6SL',
    'TILT', 'KEYWORD', 'ALFX', 'ALFY', 'BETX', 'BETY',
    'VKICK', 'HKICK', 'E1', 'E2', 'FINT', 'FINTX', 'HGAP'])

_ignoreableThinElements = {"MONITOR", "PLACEHOLDER", "MARKER",
                           "RCOLLIMATOR", "ECOLLIMATOR",
                           "COLLIMATOR", "INSTRUMENT"}

# Constants
# anything below this length is treated as a thin element
_THIN_ELEMENT_THRESHOLD = 1e-6


def ZeroMissingRequiredColumns(tfsinstance):
    """
    Sets any missing required columns to zero.  Warns user when doing so.
    """
    missingColumns = [key for key
                      in _requiredKeys if key not in tfsinstance.columns]

    if not missingColumns:
        return

    for column in missingColumns:
        tfsinstance.columns.append(column)
        for key, data in tfsinstance.data.items():
            data.append(0.0)

    missingColsString = ", ".join(["\"{}\"".format(col)
                                   for col in missingColumns])
    msg = ("Columns missing from TFS: {}.  All have been set"
           " to zero.").format(missingColsString)
    print(msg)

def _WillIgnoreItem(item, tfsinstance, ignoreZeroLength, ignoreableThinElements):
    tresult    = item['KEYWORD'] in ignoreableThinElements    
    zerolength = item['L'] < 1e-9
    result     = not tfsinstance.ElementPerturbs(item) and zerolength and ignoreZeroLength and tresult
    return result

def MadxTfs2Gmad(tfs, outputfilename,
                 startname             = None,
                 stopname              = None,
                 stepsize              = 1,
                 ignorezerolengthitems = True,
                 samplers              = 'all',
                 aperturedict          = {},
                 aperlocalpositions    = {},
                 collimatordict        = {},
                 userdict              = {},
                 partnamedict          = {},
                 verbose               = False,
                 beam                  = True,
                 flipmagnets           = None,
                 usemadxaperture       = False,
                 defaultAperture       = 'circular',
                 biases                = None,
                 allelementdict        = {},
                 optionsdict           = {},
                 beamparamsdict        = {},
                 linear                = False,
                 overwrite             = True,
                 write                 = True,
                 allNamesUnique        = False,
                 namePrepend           = ""):
    """
    **MadxTfs2Gmad** convert a madx twiss output file (.tfs) into a gmad tfs file for bdsim

    Example:

    >>> a,b = pybdsim.Convert.MadxTfs2Gmad('twiss.tfs', 'mymachine')

    returns Machine, [omittedItems]

    Returns two pybdsim.Builder.Machine instances. The first desired full conversion.  The second is
    the raw conversion that's not split by aperture. Thirdly, a list of the names of the omitted items
    is returned.

    +-------------------------------+-------------------------------------------------------------------+
    | **tfs**                       | path to the input tfs file or pymadx.Data.Tfs instance            |
    +-------------------------------+-------------------------------------------------------------------+
    | **outputfilename**            | requested output file                                             |
    +-------------------------------+-------------------------------------------------------------------+
    | **startname**                 | the name (exact string match) of the lattice element to start the |
    |                               | machine at this can also be an integer index of the element       |
    |                               | sequence number in madx tfs. This item is included in the lattice |
    +-------------------------------+-------------------------------------------------------------------+
    | **stopname**                  | the name (exact string match) of the lattice element to stop the  |
    |                               | machine at this can also be an integer index of the element       |
    |                               | sequence number in madx tfs. This item is not included            |
    +-------------------------------+-------------------------------------------------------------------+
    | **stepsize**                  | the slice step size. Default is 1, but -1 also useful for         |
    |                               | reversed line.                                                    |
    +-------------------------------+-------------------------------------------------------------------+
    | **ignorezerolengthitems**     | nothing can be zero length in bdsim as real objects of course     |
    |                               | have some finite size.  Markers, etc are acceptable but for large |
    |                               | lattices this can slow things down. True allows to ignore these   |
    |                               | altogether, which doesn't affect the length of the machine.       |
    +-------------------------------+-------------------------------------------------------------------+
    | **samplers**                  | can specify where to set samplers - options are None, 'all', or a |
    |                               | list of names of elements (normal python list of strings). Note   |
    |                               | default 'all' will generate separate outputfilename_samplers.gmad |
    |                               | with all the samplers which will be included in the main .gmad    |
    |                               | file - you can comment out the include to therefore exclude all   |
    |                               | samplers and retain the samplers file.                            |
    +-------------------------------+-------------------------------------------------------------------+
    | **aperturedict**              | Aperture information. Can either be a dictionary of dictionaries  |
    |                               | with the the first key the exact name of the element and the      |
    |                               | daughter dictionary containing the relevant bdsim parameters as   |
    |                               | keys (must be valid bdsim syntax). Alternatively, this can be a   |
    |                               | pymadx.Aperture instance that will be queried.                    |
    +-------------------------------+-------------------------------------------------------------------+
    | **aperlocalpositions**        | Dictionary of element indices to local aperture definitions       |
    |                               | of the form                                                       |
    |                               | {1: [(0.0, {"APERTYPE": "CIRCULAR", "APER1": 0.4}),               |
    |                               |      (0.5, {"APERTYPE": "ELLIPSE",  "APER1": 0.3, "APER2": 0.4}), |
    |                               |      ...],                                                        |
    |                               |  2: [...],                                                        |
    |                               | }                                                                 |
    |                               | This defines apertures in the element at index 1                  |
    |                               | starting with a CIRCULAR aper from 0.0m (i.e. the start) before   |
    |                               | changing to ELLIPSE 0.5m into the element, with possible further  |
    |                               | changes not displayed above.  As the aperture definition in GMAD  |
    |                               | is tied inseparable from its aperture definition, and vice versa, |
    |                               | this conversion function will automatically split the element     |
    |                               | at the provided local aperture points whilst retaining optical    |
    |                               | correctness.                                                      |
    |                               | This kwarg is mutually exclusive with 'aperturedict'.             |
    +-------------------------------+-------------------------------------------------------------------+
    | **collimatordict**            | A dictionary of dictionaries with collimator information keys     |
    |                               | should be exact string match of element name in tfs file value    |
    |                               | should be dictionary with the following keys:                     |
    |                               | 'bdsim_material'   - the material                                 |
    |                               | 'angle'            - rotation angle of collimator in radians      |
    |                               | 'xsize'            - x full width in metres                       |
    |                               | 'ysize'            - y full width in metres                       |
    +-------------------------------+-------------------------------------------------------------------+
    | **userdict**                  | A python dictionary the user can supply with any additional       |
    |                               | information for that particular element. The dictionary should    |
    |                               | have keys matching the exact element name in the Tfs file and     |
    |                               | contain a dictionary itself with key, value pairs of parameters   |
    |                               | and values to be added to that particular element.                |
    +-------------------------------+-------------------------------------------------------------------+
    | **partnamedict**              | A python dictionary of dictionaries. The key is a substring of    |
    |                               | that should be matched. ie add the parameter 'vhRatio' : 1 to all |
    |                               | elements with 'MBVA' in their name.                               |
    +-------------------------------+-------------------------------------------------------------------+
    | **verbose**                   | Print out lots of information when building the model.            |
    +-------------------------------+-------------------------------------------------------------------+
    | **beam**                      | True \| False - generate an input gauss Twiss beam based on the   |
    |                               | values of the twiss parameters at the beginning of the lattice    |
    |                               | (startname) NOTE - we thoroughly recommend checking these         |
    |                               | parameters and this functionality is only for partial convenience |
    |                               | to have a model that works straight away.                         |
    +-------------------------------+-------------------------------------------------------------------+
    | **flipmagnets**               | True \| False - flip the sign of all k values for magnets - MADX  |
    |                               | currently tracks particles agnostic of the particle charge -      |
    |                               | BDSIM however, follows the definition strictly -                  |
    |                               | positive k -> horizontal focussing for positive particles         |
    |                               | therefore, positive k -> vertical focussing for negative          |
    |                               | particles. Use this flag to flip the sign of all magnets.         |
    +-------------------------------+-------------------------------------------------------------------+
    | **usemadxaperture**           | True \| False - use the aperture information in the TFS file if   |
    |                               | APER_1 and APER_2 columns exist.  Will only set if they're        |
    |                               | non-zero.  Supercedes kwargs `aperturedict` and                   |
    |                               | `aperlocalpositions`.                                             |
    +-------------------------------+-------------------------------------------------------------------+
    | **defaultAperture**           | The default aperture model to assume if none is specified.        |
    +-------------------------------+-------------------------------------------------------------------+
    | **biases**                    | Optional list of bias objects to be defined in own _bias.gmad     |
    |                               | file.  These can then be attached either with allelementdict for  |
    |                               | all components or userdict for individual ones.                   |
    +-------------------------------+-------------------------------------------------------------------+
    | **allelementdict**            | Dictionary of parameter/value pairs to be written to all          |
    |                               | components.                                                       |
    +-------------------------------+-------------------------------------------------------------------+
    | **optionsdict**               | Optional dictionary of general options to be written to the       |
    |                               | bdsim model options.                                              |
    +-------------------------------+-------------------------------------------------------------------+
    | **beamparamsdict**            | Optional dictionary of parameters to be passed to the beam.       |
    +-------------------------------+-------------------------------------------------------------------+
    | **linear**                    | Only linear optical components                                    |
    +-------------------------------+-------------------------------------------------------------------+
    | **overwrite**                 | Do not append an integer to the base file name if it already      |
    |                               | exists.  Instead overwrite the files.                             |
    +-------------------------------+-------------------------------------------------------------------+
    | **allNamesUnique**            | Treat every row in the TFS file/instance as a unique element.     |
    |                               | This makes it easier to edit individual components as they are    |
    |                               | guaranteed to appear only once in the entire resulting GMAD       |
    |                               | lattice.                                                          |
    +-------------------------------+-------------------------------------------------------------------+
    | **write**                     | Whether to write the converted machine to file or not.            |
    +-------------------------------+-------------------------------------------------------------------+
    | **namePrepend**               | Optional string prepended to the name of every component.         |
    +-------------------------------+-------------------------------------------------------------------+

    """

    # machine instance that will be added to
    machine = _Builder.Machine()

    # test whether filepath or tfs instance supplied
    madx = _pymadx.Data.CheckItsTfs(tfs)

    # not very elegant but needs to be done
    varnames = ['collimatordict','userdict','partnamedict','allelementdict','optionsdict','beamparamsdict']
    vars     = [collimatordict,   userdict,  partnamedict,  allelementdict,  optionsdict,  beamparamsdict]
    for var,varname in zip(vars,varnames):
        typevar = type(var)
        if typevar not in (dict, _Data.BDSAsciiData):
            raise TypeError("Argument '" + varname + "' is not a dictionary")

    if usemadxaperture:
        aperturedict = madx
    elif aperturedict and aperlocalpositions:
        msg = "'aperturedict' and 'aperlocalpositions' are mutually exclusive."
        raise TypeError(msg)

    if "PARTICLE" in madx.header and flipmagnets is None:
        # try to check automatically
        particleName = madx.header['PARTICLE']
        if particleName == "ELECTRON":
            flipmagnets = True
            print('Detected electron in TFS file - changing flipmagnets to True')

    # If we have collimators but no collimator dict then inform that
    # they will be converted to drifts.  should really check
    # tfs[startname..]
    if "APERTYPE" in madx.columns:
        if (("RCOLLIMATOR" in madx.GetColumn("APERTYPE")
             or "ECOLLIMATOR" in madx.GetColumn("APERTYPE"))
                and not collimatordict):
            _warnings.warn("No collimatordict provided.  ALL collimators will be converted to DRIFTs.")

    if biases is not None:
        machine.AddBias(biases)

    # check whether it has all the required columns.
    ZeroMissingRequiredColumns(madx)

    if verbose:
        madx.ReportPopulations()

    # keep list of omitted zero length items
    itemsomitted = []

    # iterate through input file and construct machine
    for item in madx[startname:stopname:stepsize]:
        name = item['NAME']
        t = item['KEYWORD']
        l = item['L']
        i = item['INDEX']

        zerolength = True if item['L'] < 1e-9 else False
        if (_WillIgnoreItem(item, madx, ignorezerolengthitems, _ignoreableThinElements)):
            if verbose:
                print('skipping zero-length item: {}'.format(name))
            itemsomitted.append(name)
            continue  # skip this item.

        gmadElement = _Tfs2GmadElementFactory(item, allelementdict, verbose,
                                              userdict, collimatordict, partnamedict,
                                              flipmagnets, linear,
                                              zerolength, ignorezerolengthitems,
                                              allNamesUnique,
                                              namePrepend)
        if gmadElement is None: # factory returned nothing, go to next item.
            continue
        # We generally convert unsupported elements in the factory to
        # drifts, but if that then results in a drift with zero
        # length, then skip it.
        elif gmadElement.length == 0.0 and isinstance(gmadElement,
                                                      _Builder.Drift):
            continue
        elif l == 0.0: # Don't try to attach apertures to thin elements.
            machine.Append(gmadElement)
        elif name in collimatordict: # Don't add apertures to collimators
            machine.Append(gmadElement)
        elif aperlocalpositions: # split aperture if provided.
            elements_split_with_aper = _GetElementSplitByAperture(gmadElement,
                                                                  aperlocalpositions[i])
            for ele in elements_split_with_aper:
                machine.Append(ele)
        else: # Get element with single aperture
            element_with_aper = _GetSingleElementWithAper(item,
                                                          gmadElement,
                                                          aperturedict,
                                                          defaultAperture)
            machine.Append(element_with_aper)

    if (samplers is not None):
        machine.AddSampler(samplers)

    # Make beam file
    if beam:
        bm = MadxTfs2GmadBeam(madx, startname, verbose, extraParamsDict=beamparamsdict)
        machine.AddBeam(bm)

    options = _Options()
    if optionsdict:
        options.update(optionsdict)  # expand with user supplied bdsim options
    machine.AddOptions(options)

    if verbose:
        print('Total length: ', machine.GetIntegratedLength())
        print('Total angle:  ', machine.GetIntegratedAngle())
        print('items omitted: ')
        print(itemsomitted)
        print('number of omitted items: ', len(itemsomitted))

    if write:
        machine.Write(outputfilename, overwrite=overwrite)
    # We return machine twice to not break old interface of returning
    # two machines.
    return machine, itemsomitted


def _Tfs2GmadElementFactory(item, allelementdict, verbose,
                            userdict, collimatordict, partnamedict, flipmagnets,
                            linear, zerolength,
                            ignorezerolengthitems,
                            allNamesUnique,
                            namePrepend=""):
    """
    Function which makes the correct GMAD element given a TFS
    element to gmad.
    """
    factor = 1
    if flipmagnets is not None:
        factor = -1 if flipmagnets else 1  # flipping magnets

    # if it's already a prepared element, just return it
    if isinstance(item, _Builder.Element):
        return item

    kws = {}  # ensure empty
    # deep copy as otherwise allelementdict gets irreparably changed!
    kws = _deepcopy(allelementdict)
    if verbose:
        print('Starting key word arguments from all element dict')
        print(kws)

    name = item['NAME']
    # remove special characters like $, % etc 'reduced' name - rname:
    rname = pybdsim._General.PrepareReducedName(name
                                                if not allNamesUnique
                                                else item["UNIQUENAME"])
    t = item['KEYWORD']
    l = item['L']
    tilt = item['TILT']

    if tilt != 0:
        kws['tilt'] = tilt

    # append any user defined parameters for this element into the
    # kws dictionary

    rname = namePrepend + rname
    # name appears in the madx.  try this first.
    if name in userdict:
        kws.update(userdict[name])
    elif rname in userdict:  # rname appears in the gmad
        kws.update(userdict[rname])

    for partname in partnamedict:
        if partname in name:
            kws.update(partnamedict[partname])

    if verbose:
        print('Full set of key word arguments:')
        print(kws)

    if t == 'DRIFT':
        return _Builder.Drift(rname, l, **kws)
    elif t == 'HKICKER':
        if verbose:
            print('HICKER', rname)
        hkick = item['HKICK'] * factor
        if not zerolength:
            if l > _THIN_ELEMENT_THRESHOLD:
                kws['l'] = l
        return _Builder.HKicker(rname, hkick=hkick, **kws)
    elif t == 'VKICKER':
        if verbose:
            print('VKICKER', rname)
        vkick = item['VKICK'] * factor
        if not zerolength:
            if l > _THIN_ELEMENT_THRESHOLD:
                kws['l'] = l
        return _Builder.VKicker(rname, vkick=vkick, **kws)
    elif t == 'KICKER':
        if verbose:
            print('KICKER', rname)
        hkick = item['HKICK'] * factor
        vkick = item['VKICK'] * factor
        if not zerolength:
            if l > _THIN_ELEMENT_THRESHOLD:
                kws['l'] = l
        return _Builder.Kicker(rname, hkick=hkick, vkick=vkick, **kws)
    elif t == 'TKICKER':
        if verbose:
            print('TKICKER', rname)
        hkick = item['HKICK'] * factor
        vkick = item['VKICK'] * factor
        if not zerolength:
            if l > _THIN_ELEMENT_THRESHOLD:
                kws['l'] = l
        return _Builder.TKicker(rname, hkick=hkick, vkick=vkick, **kws)
    elif t == 'INSTRUMENT':
        # most 'instruments' are just markers
        if zerolength and not ignorezerolengthitems:
            return _Builder.Marker(rname)
        return _Builder.Drift(rname, l, **kws)
    elif t == 'MARKER':
        if not ignorezerolengthitems:
            return _Builder.Marker(rname)
    elif t == 'MONITOR':
        # most monitors are just markers
        if zerolength and not ignorezerolengthitems:
            return _Builder.Marker(rname)
        return _Builder.Drift(rname, l, **kws)
    elif t == 'MULTIPOLE':
        k1 = item['K1L'] * factor
        k2 = item['K2L'] * factor if not linear else 0
        k3 = item['K3L'] * factor if not linear else 0
        k4 = item['K4L'] * factor if not linear else 0
        k5 = item['K5L'] * factor if not linear else 0
        k6 = item['K6L'] * factor if not linear else 0
        k1s = item['K1SL'] * factor
        k2s = item['K2SL'] * factor if not linear else 0
        k3s = item['K3SL'] * factor if not linear else 0
        k4s = item['K4SL'] * factor if not linear else 0
        k5s = item['K5SL'] * factor if not linear else 0
        k6s = item['K6SL'] * factor if not linear else 0

        knl = (k1,  k2,  k3,  k4,  k5,  k6)
        ksl = (k1s, k2s, k3s, k4s, k5s, k6s)

        finiteStrength = _np.any(
            [k1, k2, k3, k4, k5, k6, k1s, k2s, k3s, k4s, k5s, k6s])

        if zerolength or l < _THIN_ELEMENT_THRESHOLD:
            return _Builder.ThinMultipole(rname, knl=knl, ksl=ksl, **kws)
        else:
            return _Builder.Multipole(rname, l, knl=knl, ksl=ksl, **kws)
    elif t == 'OCTUPOLE':
        if zerolength or l < _THIN_ELEMENT_THRESHOLD:
            k3 = item['K3L'] * factor if not linear else 0
            return _Builder.ThinMultipole(rname, knl=(0, 0, k3), **kws)
        k3 = item['K3L'] / l * factor if not linear else 0
        return _Builder.Octupole(rname, l, k3=k3, **kws)
    elif t == 'PLACEHOLDER':
        if zerolength:
            if not ignorezerolengthitems:
                return _Builder.Marker(rname)
        else:
            return _Builder.Drift(rname, l, **kws)
    elif t == 'QUADRUPOLE':
        if zerolength or l < _THIN_ELEMENT_THRESHOLD:
            k1 = item['K1L'] * factor
            return _Builder.ThinMultipole(rname, knl=(k1,), **kws)
        k1 = item['K1L'] / l * factor
        return _Builder.Quadrupole(rname, l, k1, **kws)
    elif t == 'RBEND':
        angle = item['ANGLE']
        e1 = item['E1'] if 'E1' in item else 0
        e2 = item['E2'] if 'E2' in item else 0
        fint = item['FINT'] if 'FINT' in item else 0
        fintx = item['FINTX'] if 'FINTX' in item else -1 # madx convention -> -1 => same as FINT
        h1 = item['H1'] if 'H1' in item else 0
        h2 = item['H2'] if 'H2' in item else 0
        hgap = item['HGAP'] if 'HGAP' in item else 0
        k1l = item['K1L']
        # set element length to be the chord length - tfs output rbend
        # length is arc length
        chordLength = l
        if angle != 0:
            chordLength = 2 * (l / angle) * _np.sin(angle / 2.) #protect against 0 angle rbends
        # subtract dipole angle/2 added on to poleface angles internally by
        # madx
        poleInAngle = e1 - 0.5 * angle
        poleOutAngle = e2 - 0.5 * angle
        if poleInAngle != 0:
            kws['e1'] = poleInAngle
        if poleOutAngle != 0:
            kws['e2'] = poleOutAngle
        if fint != 0:
            kws['fint'] = fint
        # in madx, -1 means fintx was allowed to default to fint and we should do the same
        # so if set to 0, this means we want it to be 0
        if fintx != -1:
            kws['fintx'] = fintx
        if h1 != 0:
            kws['h1'] = h1
        if h2 != 0:
            kws['h2'] = h2
        if hgap != 0:
            kws['hgap'] = hgap
        if k1l != 0:
            # NOTE we don't use factor here for magnet flipping
            k1 = k1l / l
            kws['k1'] = k1
        return _Builder.RBend(rname, chordLength, angle=angle, **kws)
    elif t == 'SBEND':
        angle = item['ANGLE']
        e1 = item['E1'] if 'E1' in item else 0
        e2 = item['E2'] if 'E2' in item else 0
        fint = item['FINT'] if 'FINT' in item else 0
        fintx = item['FINTX'] if 'FINTX' in item else -1
        h1 = item['H1'] if 'H1' in item else 0
        h2 = item['H2'] if 'H2' in item else 0
        hgap = item['HGAP'] if 'HGAP' in item else 0
        k1l = item['K1L']
        if e1 != 0:
            kws['e1'] = e1
        if e2 != 0:
            kws['e2'] = e2
        if k1l != 0:
            # NOTE we're not using factor for magnet flipping here
            k1 = k1l / l
            kws['k1'] = k1

        # if fint != 0:
        kws['fint'] = fint

        # in madx, -1 means fintx was allowed to default to fint and we should do the same
        # so if set to 0, this means we want it to be 0
        if fintx == -1:
            if fint:
                kws['fintx'] = fint
            else:
                kws['fintx'] = 0
        else:
            kws['fintx'] = fintx
        if h1 != 0:
            kws['h1'] = h1
        if h2 != 0:
            kws['h2'] = h2
        # if hgap != 0:
        kws['hgap'] = hgap
        return _Builder.SBend(rname, l, angle=angle, **kws)
    elif t in {'RCOLLIMATOR', 'ECOLLIMATOR', 'COLLIMATOR'}:
        # only use xsize as only have half gap
        if name in collimatordict:
            # gets a dictionary then extends kws dict with that dictionary
            colld = collimatordict[name]

            # collimator defined by external geometry file
            if 'geometryFile' in colld:
                kws['geometryFile'] = colld['geometryFile']
                k = 'outerDiameter'  # key
                if k not in kws:
                    # not already specified via other dictionaries
                    if k in colld:
                        kws[k] = colld[k]
                    else:
                        kws[k] = 1  # ensure there's a default as not in madx
                # add a general element
                return _Builder.ExternalGeometry(rname, l, **kws)
            else:
                kws['material'] = colld.get('material', 'copper')
                tilt = colld.get('tilt', 0)
                if tilt != 0:
                    kws['tilt'] = tilt
                try:
                    xsize = colld['xsize']
                    ysize = colld['ysize']
                except KeyError:
                    xsize = colld['XSIZE']
                    ysize = colld['YSIZE']
                if verbose:
                    print('collimator x,y size ', xsize, ysize)
                if 'outerDiameter' in colld:
                    kws['outerDiameter'] = colld['outerDiameter']
                else:
                    kws['outerDiameter'] = max([0.5,
                                                xsize * 2.5,
                                                ysize * 2.5])
                if t == 'RCOLLIMATOR' or t == "COLLIMATOR":
                    return _Builder.RCol(rname, l, xsize, ysize, **kws)
                else:
                    return _Builder.ECol(rname, l, xsize, ysize, **kws)
        # dict is incomplete or the component is erroneously
        # reffered to as a collimator even when it can be thought
        # of as a drift (e.g. LHC TAS).
        elif collimatordict != {}:
            msg = ("{} {} not found in collimatordict."
                   " Will instead convert to a DRIFT!  This is not"
                   " necessarily wrong!".format(t, name))
            _warnings.warn(msg)
            return _Builder.Drift(rname, l, **kws)
        # if user didn't provide a collimatordict at all.
        else:
            return _Builder.Drift(rname, l, **kws)
    elif t == 'RFCAVITY':
        return _Builder.Drift(rname, l, **kws)
    elif t == 'SEXTUPOLE':
        if zerolength or l < _THIN_ELEMENT_THRESHOLD:
            k2 = item['K2L'] * factor if not linear else 0
            return _Builder.ThinMultipole(rname, knl=(0, k2), **kws)
        k2 = item['K2L'] / l * factor if not linear else 0
        return _Builder.Sextupole(rname, l, k2, **kws)
    elif t == 'SOLENOID':
        if l == 0.0:
            msg = "{} is a thin solenoid and is omitted.".format(name)
            _warnings.warn(msg)
            return None
        ks = item['KSI'] / l
        return _Builder.Solenoid(rname, l, ks=ks, **kws)
    else:
        print('unknown element type:', t, 'for element named: ', name)
        if zerolength and not ignorezerolengthitems:
            print('putting marker in instead as its zero length')
            return _Builder.Marker(rname)
        print('putting drift in instead as it has a finite length')
        return _Builder.Drift(rname, l)

    raise ValueError("Unable to construct Element: {}, {}".format(t, name))

def _GetElementSplitByAperture(gmadElement, localApertures):
    # tolerate any bad apertures - like only a few - and just don't append them
    apertures = []
    for point, aper in localApertures:
        try:
            arp = _Builder.PrepareApertureModel(aper, warningName=gmadElement.name)
            apertures.append(arp)
        except ValueError:
            pass
        
    if localApertures[0][0] != 0.0 :
        raise ValueError("No aperture defined at start of element.")
    if len(localApertures) > 1:
        split_points = [point for point, _  in localApertures[1:]]
        split_elements = gmadElement.split(split_points)
        for aper, split_element in zip(apertures, split_elements):
            split_element.update(aper)
        return split_elements
    elif len(localApertures) == 1:
        gmadElement = _deepcopy(gmadElement)
        gmadElement.update(apertures[0])
        return [gmadElement]
    raise ValueError("Unable to split element by apertures.")

def _GetSingleElementWithAper(item, gmadElement,
                              aperturedict, defaultAperture):
    """Returns the raw aperture model (i.e. unsplit), and a list of
    split apertures."""
    gmadElement = _deepcopy(gmadElement)
    name = item["NAME"]
    # note SORIGINAL not S.  This is so it works still after slicing.
    sMid = item["SORIGINAL"] - item["L"] / 2.0
    aper = {}
    try:
        aper = _Builder.PrepareApertureModel(
            aperturedict.GetApertureAtS(sMid), defaultAperture)
    except AttributeError:
        pass
    try:
        this_aperdict = aperturedict[name]
    except KeyError:
        pass
    else:
        aper = _Builder.PrepareApertureModel(this_aperdict, defaultAperture)

    gmadElement.update(aper)
    return gmadElement


def MadxTfs2GmadBeam(tfs, startname=None, verbose=False, extraParamsDict={}):
    """
    Takes a pymadx.Data.Tfs instance and extracts information from first line to
    create a BDSIM beam definition in a pybdsim.Beam object.  Note
    that if kwarg startname is used, the optics are retrieved at the
    start of the element, i.e. you do not need to get the optics of
    the previous element, this function does that automatically.

    Works for e+, e- and proton.
    Default emittance is 1e-9mrad if 1 in tfs file.


    """
    print('Warning - using automatic generation of input beam distribution from madx tfs file - PLEASE CHECK!')

    if startname is None:
        startindex = 0
    else:
        try:
            startindex = tfs.IndexFromName(startname)
        except ValueError:  # Then assume it's already an index.
            startindex = startname

    # MADX defines parameters at the end of elements so need to go 1 element
    # back if we can.

    if startindex > 0:
        startindex -= 1

    energy = float(tfs.header['ENERGY'])
    particle = tfs.header['PARTICLE']
    ex = tfs.header['EX']
    ey = tfs.header['EY']
    if 'EX' in extraParamsDict:
        ex = extraParamsDict['EX']
    if 'EY' in extraParamsDict:
        ey = extraParamsDict['EY']
    sigmae = float(tfs.header['SIGE'])

    if ex == 1:
        print('Horizontal emittance of 1 is too large - setting to 1e-9')
        ex = 1e-9
    if ey == 1:
        print('Horizontal emittance of 1 is too large - setting to 1e-9')
        ey = 1e-9

    data = tfs[startindex]

    if particle == 'ELECTRON':
        particle = 'e-'
    elif particle == 'POSITRON':
        particle = 'e+'
    elif particle == 'PROTON':
        particle = 'proton'
    else:
        raise ValueError("Unsupported particle " + particle)

    if verbose:
        print('beta_x: ', data['BETX'], 'alpha_x: ', data['ALFX'], 'mu_x: ', data['MUX'])
        print('beta_y: ', data['BETY'], 'alpha_y: ', data['ALFY'], 'mu_y: ', data['MUY'])

    # note, in the main pybdsim.__init__.py Beam class is imported from Beam.py
    # so in this submodule when we do from .. import Beam it's actually the
    # already imported class that's being imported
    beam = _Beam.Beam(particle, energy, 'gausstwiss')
    if ex != 0:
        beam.SetEmittanceX(ex, 'm')
    if ey != 0:
        beam.SetEmittanceY(ey, 'm')
    beam.SetSigmaE(sigmae)

    beamparams = {"SetBetaX":'BETX',
                  "SetBetaY":'BETY',
                  "SetAlphaX":'ALFX',
                  "SetAlphaY":'ALFY',
                  "SetDispX":'DXBETA',
                  "SetDispY":'DYBETA',
                  "SetDispXP":'DPXBETA',
                  "SetDispYP":'DPYBETA',
                  "SetXP0": 'PX',
                  "SetYP0": 'PY',
                  "SetX0": 'X',
                  "SetY0": 'Y'
                  }
    for func,parameter in beamparams.items():
        if parameter in list(data.keys()):
            getattr(beam, func)(data[parameter])

    for k, v in extraParamsDict.items():
        beam[k] = v

    return beam



import os.path
from itertools import islice

import pytest
import pybdsim
import pymadx

from . import pybdsim_test_utils as utils

# TODO:
# TEST with aperturedict of pymadx.Data.Aperture.
# TEST with collimator dict (atf2 has no collimators).
# TEST with usemadxaperture (atf2 has no aperinfo with it).
# TEST coupled combinations of arguments, whatever they may be.

PATH_TO_TEST_INPUT = "{}/../test_input/".format(os.path.dirname(__file__))
PATH_TO_TEST_OUTPUT = "{}/../test_output/".format(os.path.dirname(__file__))

GMAD_SUFFIXES = ["",
                 "_beam",
                 "_components",
                 "_options",
                 "_sequence",
#                 "_bias",
#                 "_samplers"
]


@pytest.fixture
def atf2_input():
    return "{}/atf2-nominal-twiss-v5.2.tfs.tar.gz".format(
        PATH_TO_TEST_INPUT)

@pytest.fixture
def model_model_input(tmppath):
    tfs = "{}/model-model.tfs.gz".format(PATH_TO_TEST_INPUT)
    aper = pymadx.Data.Aperture("{}/model-model-aper.tfs.gz".format(PATH_TO_TEST_INPUT))
    aper = aper.RemoveBelowValue(0.005)
    coll = pybdsim.Data.Load("{}/model-model-collsettings.dat".format(PATH_TO_TEST_INPUT))
    return {"tfs": tfs, "outputfilename": tmppath, "aperturedict": aper, "collimatordict": coll}

@pytest.fixture
def tmppath(tmpdir):
    """A temporary file path"""
    return str(tmpdir.mkdir("testdir").join("model"))

@pytest.fixture(params=["one", "list"]) # bias or list of biases..
def biases(request):
    """Biases can be either a single XSecBias instance or a list
    thereof.  This fixture provides both."""
    bias1 = pybdsim.XSecBias.XSecBias("mydecay1", "gamma", "decay", "1e5", "2")
    bias2 = pybdsim.XSecBias.XSecBias("mydecay2", "proton", "decay", "1e5", "2")
    if request.param == "bias":
        return bias1
    if request.param == "list":
        return [bias1, bias2]

@pytest.fixture
def gmad_comparator():
    def f(first, second):
        """Paths to two sets of gmad files"""
        for suffix in GMAD_SUFFIXES:
            print("comparing '"+suffix+"'")
            new_gmad = "{}{}.gmad".format(first, suffix)
            old_gmad = "{}{}.gmad".format(second, suffix)
            # Skip the first 3 lines which are just the header
            with open(new_gmad, "r") as new, open(old_gmad, "r") as old:
                for new_line, old_line in zip(islice(new, 3, None),
                                              islice(old, 3, None)):
                    assert new_line == old_line
    return f

@pytest.mark.sanity
def test_atf2_conversion_default(atf2_input, tmppath):
    """Default parameters should not fail."""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath)

@pytest.mark.sanity
@pytest.mark.parametrize(('start', 'stop', 'step'),
                         [(10, 20, 2),
                          ("KEX1A", "L229", 1),
                          ("KEX1A", 40, 1),
                          (10, "L229", 1)])
def test_atf2_conversion_with_start_stop_and_stepsize(atf2_input, tmppath,
                                                      start, stop, step):
    """Given the ATF2 model and a start, stop and step:  do not crash."""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath,
                                 startname=start,
                                 stopname=stop,
                                 stepsize=step)
@pytest.mark.sanity
@pytest.mark.parametrize('ignorezerolengthitems', [True, False])
def test_atf2_conversion_with_ignorezerolengthitems(atf2_input, tmppath,
                                                    ignorezerolengthitems):
    """Given the ATF2 model and valid args for
    ignorezerolengthitems:  do not crash."""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath,
                                 ignorezerolengthitems=ignorezerolengthitems)

@pytest.mark.sanity
@pytest.mark.parametrize('flipmagnets', [True, False, None])
def test_atf2_conversion_with_flipmagnets(atf2_input, tmppath, flipmagnets):
    """Given the ATF2 model and the set of allowed `flipmagnets` arguments:
    do not crash."""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, flipmagnets=flipmagnets)

@pytest.mark.sanity
@pytest.mark.parametrize('linear', [True, False])
def test_atf2_conversion_with_linear(atf2_input, tmppath, linear):
    """Given the ATF2 model and the set of allowed `linear` arguments:
    do not crash."""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, linear=linear)

@pytest.mark.sanity
@pytest.mark.parametrize('samplers', ['all', None, ["KEX1A", "KEX1B"]])
def test_atf2_conversion_with_samplers(atf2_input, tmppath, samplers):
    """Given the ATF2 model and the set of allowed `samplers` arguments:
    do not crash."""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, samplers=samplers)


@pytest.mark.sanity
def test_atf2_conversion_with_aperturedict(atf2_input, tmppath):
    # should also test with a pymadx.Data.Aperture instance.
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath,
                                 aperturedict={
                                     "KEX1A":
                                     {"APERTYPE": "circular",
                                      "APER_1": 1,
                                      "APER_2": 0,
                                      "APER_3": 0,
                                      "APER_4": 0}})

@pytest.mark.sanity
def test_atf2_conversion_with_optionsDict(atf2_input, tmppath):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, optionsdict={"stopSecondaries": "1"})

@pytest.mark.sanity
def test_atf2_conversion_with_userdict(atf2_input, tmppath):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, userdict={"KEX1A": {"biasVacuum": "mybias"}})

@pytest.mark.sanity
def test_atf2_conversion_with_allelementdict(atf2_input, tmppath):
    """Don't crash for valid arguments of allelementdict"""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, allelementdict={"biasVacuum": "mybias"})

@pytest.mark.sanity
def test_atf2_conversion_with_defaultAperture(atf2_input, tmppath):
    """Don't crash for valid arguments of defaultAperture"""
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, defaultAperture="rectellipse")

@pytest.mark.sanity
def test_atf2_conversion_with_biases(atf2_input, tmppath, biases):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, biases=biases)

@pytest.mark.parametrize('beam', [True, False])
def test_atf2_conversion_with_beam(atf2_input, tmppath, beam):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, beam=beam)

@pytest.mark.parametrize('overwrite', [True, False])
@pytest.mark.sanity
def test_atf2_conversion_with_overwrite(atf2_input, tmppath, overwrite):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, overwrite=overwrite)

@pytest.mark.parametrize('allNamesUnique', [True, False])
@pytest.mark.sanity
def test_atf2_conversion_with_allNamesUnique(atf2_input, tmppath, allNamesUnique):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, allNamesUnique=allNamesUnique)

@pytest.mark.parametrize('verbose', [True, False])
@pytest.mark.sanity
def test_atf2_conversion_with_verbose(atf2_input, tmppath, verbose):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, verbose=verbose)

@pytest.mark.sanity
def test_atf2_conversion_with_beamParmsDict(atf2_input, tmppath):
    beam = pybdsim.Convert.MadxTfs2GmadBeam(pymadx.Data.Tfs(atf2_input),
                                            startname="KEX1A")
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath, beam=beam)

@pytest.mark.regression
def test_atf2_default_conversion_with_gmad_regression(atf2_input,
                                                      gmad_comparator,
                                                      tmppath):
    pybdsim.Convert.MadxTfs2Gmad(atf2_input, tmppath)
    output_path = utils.get_output_path("atf2-nominal-twiss-v5.2.tfs.tar.gz")
    gmad_comparator(output_path, tmppath)

@pytest.mark.regression
def test_model_model_conversion_with_gmad_regression(model_model_input,
                                                     gmad_comparator):
    pybdsim.Convert.MadxTfs2Gmad(**model_model_input)
    output_path = utils.get_output_path("model-model.tfs.gz")
    gmad_comparator(output_path, model_model_input['outputfilename'])



"""
Module for various conversions.

"""
# pymadx
from ._MadxTfs2Gmad import MadxTfs2Gmad
from ._MadxTfs2Gmad import ZeroMissingRequiredColumns as _ZeroMissingRequiredColumns
from ._MadxTfs2GmadStrength import MadxTfs2GmadStrength
from ._MadxTfs2Gmad import MadxTfs2GmadBeam

from ._BdsimPrimaries2Inrays import BdsimPrimaries2Ptc
from ._BdsimPrimaries2Inrays import BdsimPrimaries2Madx
from ._BdsimPrimaries2Inrays import BdsimPrimaries2Mad8
from ._BdsimPrimaries2Inrays import BdsimSampler2Ptc
from ._BdsimPrimaries2Inrays import BdsimPrimaries2BdsimUserFile
from ._BdsimPrimaries2Inrays import BdsimSampler2BdsimUserFile

from ._BdsimElement2TransferMatrix import BdsimElement2TransferMatrix

from ._ElegantParamToStrength import ElegantParam2GmadStrength

from ._Transport2Gmad import Transport2Gmad

# optional pymad8
try:
    from ._Mad8Twiss2Gmad import Mad8Twiss2Gmad
except ImportError:
    pass

# optional pysad
try:
    import pysad
    from ._SadFlat2Gmad import SadFlat2GMad
except ImportError:
    pass

# optional cpymad
try:
    from ._CPyMad2Gmad import CPyMad2Gmad
except ImportError:
    pass





import vtk
import pybdsim

def Plot3DXYZVtk(filename):

    d = pybdsim.Field.Load(filename)

    glyph3D = _fieldToVtkStructuredGrid(d)

    colors = vtk.vtkNamedColors()

    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(glyph3D.GetOutputPort())

    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetColor([0,0,0])

    renderer = vtk.vtkRenderer()
    renderWindow = vtk.vtkRenderWindow()
    renderWindow.AddRenderer(renderer)
    renderWindow.SetWindowName('OrientedGlyphs')

    renderWindowInteractor = vtk.vtkRenderWindowInteractor()
    renderWindowInteractor.SetRenderWindow(renderWindow)

    axes = vtk.vtkAxesActor()
    axes.SetTotalLength(50,50,50)

    renderer.AddActor(actor)
    renderer.AddActor(axes)
    renderer.SetBackground([1,1,1])

    renderWindow.Render()
    renderWindowInteractor.Start()

def _fieldToVtkStructuredGrid(inputData) :

    nx = inputData.header['nx']
    minx = inputData.header['xmin']
    maxx = inputData.header['xmax']
    ny = inputData.header['ny']
    miny = inputData.header['ymin']
    maxy = inputData.header['ymax']
    nz = inputData.header['nz']
    minz = inputData.header['zmin']
    maxz = inputData.header['zmax']
    strucGrid = vtk.vtkStructuredGrid()

    strucGrid.SetDimensions(nx,ny,nz)

    points = vtk.vtkPoints()
    pointValues = vtk.vtkDoubleArray()
    pointValues.SetNumberOfComponents(3)
    pointValues.SetNumberOfTuples(nx*ny*nz)

    iPoint = 0
    for i in range(0,nx) :
        for j in range(0,ny) :
            for k in range(0,nz) :
                points.InsertNextPoint(inputData.data[i,j,k,0],
                                       inputData.data[i,j,k,1],
                                       inputData.data[i,j,k,2])
                pointValues.SetTuple(iPoint,
                                     [inputData.data[i,j,k,3],
                                     inputData.data[i,j,k,4],
                                     inputData.data[i,j,k,5]])
                iPoint += 1

    strucGrid.SetPoints(points)
    strucGrid.GetPointData().SetVectors(pointValues)

    arrowSource = vtk.vtkArrowSource()

    glyph3D = vtk.vtkGlyph3D()
    glyph3D.SetSourceConnection(arrowSource.GetOutputPort())
    # glyph3D.SetVectorModeToUseVector()
    glyph3D.SetScaleModeToScaleByVector()
    glyph3D.SetInputData(strucGrid)
    glyph3D.SetScaleFactor(5)
    glyph3D.Update()

    return glyph3D



"""
Functions for comparing the optical distributions of two
BDSIM models.

Functions for plotting the individual optical functions, and an
eighth, helper function ''compare_all_optics``, to plot display all
seven in one go.
"""

import os.path as _path
import matplotlib.pyplot as _plt
import datetime as _datetime
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages

import pybdsim.Data as _Data
import pybdsim.Plot as _Plot

import pymadx as _pymadx
import pytransport as _pyt

# Predefined lists of tuples for making the standard plots,
# format = (optical_var_name, optical_var_error_name, legend_name)

_BETA = [("Beta_x", "Sigma_Beta_x", r'$\beta_{x}$'),
         ("Beta_y", "Sigma_Beta_y", r'$\beta_{y}$')]

_ALPHA = [("Alpha_x", "Sigma_Alpha_x", r"$\alpha_{x}$"),
          ("Alpha_y", "Sigma_Alpha_y", r"$\alpha_{y}$")]

_DISP = [("Disp_x", "Sigma_Disp_x", r"$D_{x}$"),
         ("Disp_y", "Sigma_Disp_y", r"$D_{y}$")]

_DISP_P = [("Disp_xp", "Sigma_Disp_xp", r"$D_{p_{x}}$"),
           ("Disp_yp", "Sigma_Disp_yp", r"$D_{p_{y}}$")]

_SIGMA = [("Sigma_x", "Sigma_Sigma_x", r"$\sigma_{x}$"),
          ("Sigma_y", "Sigma_Sigma_y", r"$\sigma_{y}$")]

_SIGMA_P = [("Sigma_xp", "Sigma_Sigma_xp", r"$\sigma_{xp}$"),
            ("Sigma_yp", "Sigma_Sigma_yp", r"$\sigma_{yp}$")]

def _parse_bdsim_input(bdsim_in, name):
   """Return bdsim_in as a BDSAsciiData instance, which should either
   be a path to a BDSIM root output file, rebdsimOptics output file,
   or a BDSAsciiData instance, and in either case, generate a
   name if None is provided, and return that as well."""
   if isinstance(bdsim_in, str):
       if not _path.isfile(bdsim_in):
           raise IOError("file \"{}\" not found!".format(bdsim_in))
       name = (_path.splitext(_path.basename(bdsim_in))[0]
               if name is None else name)
       data = _Data.Load(bdsim_in)
       if hasattr(data, 'optics'):
           data = data.optics
       elif hasattr(data, 'Optics'):
           data = data.Optics
       else:
           raise AttributeError("No optics found for BDSIM file: {}".format(bdsim_in))
       return data, name
   try:
       if isinstance(bdsim_in, _Data.RebdsimFile):
           if hasattr(bdsim_in,'optics'):
               bdsim_in = bdsim_in.optics
           elif hasattr(bdsim_in,'Optics'):
               bdsim_in = bdsim_in.Optics
           else:
               raise AttributeError("No optics found in rebdsim file.")
       name = bdsim_in.filename if name is None else name
       return bdsim_in, name
   except AttributeError:
       raise TypeError(
           "Expected Tfs input is neither a "
           "file path nor a Tfs instance: {}".format(bdsim_in))

# use closure to avoid tonnes of boilerplate code as happened with
# MadxBdsimComparison.py
def _make_plotter(plot_info_tuples, x_label, y_label, title):
    def f_out(first, second, first_name=None, second_name=None,
              survey=None, **kwargs):

        first = _pyt.Reader.GetOptics(first)

        # Get the initial N for the bdsim source
        second_nparticles = second.Npart()[0]

        plot = _plt.figure(title, **kwargs)
        # Loop over the variables in plot_info_tuples and draw the plots.
        for var, error, legend_name in plot_info_tuples:
            _plt.plot(first.GetColumn('S'),
                      first.GetColumn(var),
                      label="{}; {};".format(
                          first_name, legend_name),
                      **kwargs)
            _plt.errorbar(second.GetColumn('S'),
                          second.GetColumn(var),
                          yerr=second.GetColumn(error),
                          label="{}; {}; N = {:.1E}".format(
                              second_name, legend_name, second_nparticles),
                          capsize=3, **kwargs)


        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(y_label)
        axes.set_xlabel(x_label)
        axes.legend(loc='best')

        if survey is not None:
           try:
               _Plot.AddMachineLatticeFromSurveyToFigure(_plt.gcf(), survey)
           except IOError:
               _pymadx.Plot.AddMachineLatticeToFigure(_plt.gcf(), survey)
        _plt.show(block=False)
        return plot
    return f_out

PlotBeta   = _make_plotter(_BETA,    "S / m", r"$\beta_{x,y}$ / m",      "Beta")
PlotAlpha  = _make_plotter(_ALPHA,   "S / m", r"$\alpha_{x,y}$ / m",     "Alpha")
PlotDisp   = _make_plotter(_DISP,    "S / m", r"$D_{x,y} / m$",          "Dispersion")
PlotDispP  = _make_plotter(_DISP_P,  "S / m", r"$D_{p_{x},p_{y}}$ / m",  "Momentum_Dispersion")
PlotSigma  = _make_plotter(_SIGMA,   "S / m", r"$\sigma_{x,y}$ / m",     "Sigma")
PlotSigmaP = _make_plotter(_SIGMA_P, "S / m", r"$\sigma_{xp,yp}$ / rad", "SigmaP")


def TransportVsBDSIM(first, second, first_name=None,
                     second_name=None, survey=None, saveAll=True,
                     outputFileName=None, **kwargs):
    """
    Display all the optical function plots for the two input optics files.
    """
    second, second_name = _parse_bdsim_input(second, second_name)

    figures=[PlotBeta(first, second, first_name=first_name,
                      second_name=second_name, survey=survey, **kwargs),
             PlotAlpha(first, second, first_name=first_name,
                       second_name=second_name, survey=survey, **kwargs),
             PlotDisp(first, second, first_name=first_name,
                      second_name=second_name, survey=survey, **kwargs),
             PlotDispP(first, second, first_name=first_name,
                       second_name=second_name, survey=survey, **kwargs),
             PlotSigma(first, second, first_name=first_name,
                       second_name=second_name, survey=survey, **kwargs),
             PlotSigmaP(first, second, first_name=first_name,
                        second_name=second_name, survey=survey, **kwargs)]

    if saveAll:
       if outputFileName is not None:
           output_filename = outputFileName
       else:
           output_filename = "optics-report.pdf"
       with _PdfPages(output_filename) as pdf:
           for figure in figures:
               pdf.savefig(figure)
           d = pdf.infodict()
           d['Title'] = "{} VS {} Optical Comparison".format(first_name, second_name)
           d['CreationDate'] = _datetime.datetime.today()
       print("Written ", output_filename)



from copy import deepcopy as _deepcopy
import numpy as _np
import math as _math
import pymad8 as _m8
import warnings as _warnings
from .. import Builder as _Builder
from .. import Beam as _Beam
from .. import Data as _Data
from ..Options import Options as _Options
import pybdsim._General

_ignoreableThinElements = {"MONI", "IMON", "BLMO", "MARK", "RCOL", "ECOL", "INST", "WIRE"}


# Constants
# anything below this length is treated as a thin element
_THIN_ELEMENT_THRESHOLD = 1e-6


def Mad8Twiss2Gmad(inputfilename, outputfilename,
				   startindex            = 0,
				   endindex              = -1,
				   stepsize              = 1,
				   ignorezerolengthitems = True,
				   samplers              = 'all',
				   aperturedict          = {},
				   aperlocalpositions    = {},
				   collimatordict        = {},
				   userdict              = {},
				   partnamedict          = {},
				   verbose               = False,
				   beam                  = True,
				   flipmagnets           = None,
				   defaultAperSize       = 0.1,
				   defaultAperShape      = 'circular',
				   biases                = None,
				   materials			  = None,
				   allelementdict        = {},
				   optionsdict           = {},
				   beamparamsdict        = {},
				   linear                = False,
				   overwrite             = True,
				   write                 = True,
				   namePrepend           = ""):

	"""
	**Mad82Gmad** convert a mad8 output file into a gmad tfs file for bdsim

	Example:

	>>> a,b = pybdsim.Convert.MadxTfs2Gmad('twissfile', 'mymachine')

	returns Machine, [omittedItems]

	Returns two pybdsim.Builder.Machine instances. The first desired full conversion.  The second is
	the raw conversion that's not split by aperture. Thirdly, a list of the names of the omitted items
	is returned.

	+-------------------------------+-------------------------------------------------------------------+
	| **inputfilename**             | Path to the Mad8 input file.                                      |
	+-------------------------------+-------------------------------------------------------------------+
	| **outputfilename**            | Requested output file.                                            |
	+-------------------------------+-------------------------------------------------------------------+
	| **startindex**                | Index of the lattice element to start the machine at.             |
	|                               | This item is included in the lattice.                             |
	+-------------------------------+-------------------------------------------------------------------+
	| **endindex**                  | Index of the lattice element to stop the machine at.              |
	|                               | This item is not included in the lattice.                         |
	+-------------------------------+-------------------------------------------------------------------+
	| **stepsize**                  | Slice step size. Default is 1, but -1 also useful for             |
	|                               | reversed line.                                                    |
	+-------------------------------+-------------------------------------------------------------------+
	| **ignorezerolengthitems**     | nothing can be zero length in bdsim as real objects of course     |
	|                               | have some finite size.  Markers, etc are acceptable but for large |
	|                               | lattices this can slow things down. True allows to ignore these   |
	|                               | altogether, which doesn't affect the length of the machine.       |
	+-------------------------------+-------------------------------------------------------------------+
	| **samplers**                  | can specify where to set samplers - options are None, 'all', or a |
	|                               | list of names of elements (normal python list of strings). Note   |
	|                               | default 'all' will generate separate outputfilename_samplers.gmad |
	|                               | with all the samplers which will be included in the main .gmad    |
	|                               | file - you can comment out the include to therefore exclude all   |
	|                               | samplers and retain the samplers file.                            |
	+-------------------------------+-------------------------------------------------------------------+
	| **aperturedict**              | Aperture information. Can either be a dictionary of dictionaries  |
	|                               | with the the first key the exact name of the element and the      |
	|                               | daughter dictionary containing the relevant bdsim parameters as   |
	|                               | keys (must be valid bdsim syntax).                                |
	+-------------------------------+-------------------------------------------------------------------+
	| **aperlocalpositions**        | Dictionary of element indices to local aperture definitions       |
	|                               | of the form                                                       |
	|                               | {1: [(0.0, {"APERTYPE": "circular",   "APER_1": 0.4}),            |
	|                               |      (0.5, {"APERTYPE": "elliptical", "APER_1": 0.3,              |
	|                               |                                       "APER_2": 0.4}),            |
	|                               |      ...],                                                        |
	|                               |  2: [...],                                                        |
	|                               | }                                                                 |
	|                               | This defines apertures in the element at index 1 starting with a  |
	|                               | "circular" aperture from 0.0m (i.e. the start) before changing to |
	|                               | "elliptical" after 0.5m in the element, with possible further     |
	|                               | changes not displayed above.                                      |
	|                               | As the aperture definition in GMAD is tied inseparable from its   |
	|                               | aperture definition, and vice versa, this conversion function     |
	|                               | will automatically split the element at the local aperture points |
	|                               | whilst retaining optical correctness.                             |
	|                               | This kwarg is mutually exclusive with "aperturedict".             |
	+-------------------------------+-------------------------------------------------------------------+
	| **collimatordict**            | A dictionary of dictionaries with collimator information.         |
	|                               | Keys should be exact string match of element name in Mad8 file.   |
	|                               | Value should be dictionary with the following keys:               |
	|                               | "material"         - the material                                 |
	|                               | "tilt"             - rotation angle of collimator in radians      |
	|                               | "xsize"            - x full width in metres                       |
	|                               | "ysize"            - y full width in metres                       |
	+-------------------------------+-------------------------------------------------------------------+
	| **userdict**                  | Dictionary that supply any additional information for an element. |
	|                               | Keys should match the exact element name in the Mad8 file.        |
	|                               | Value should be dictionary itself with key, value pairs of        |
	|                               | parameters and values to be added to that particular element.     |
	+-------------------------------+-------------------------------------------------------------------+
	| **partnamedict**              | Dictionary of dictionaries. The key is a substring of             |
	|                               | that should be matched. ie add the parameter 'vhRatio' : 1 to all |
	|                               | elements with 'MBVA' in their name.                               |
	+-------------------------------+-------------------------------------------------------------------+
	| **verbose**                   | Print out lots of information when building the model.            |
	+-------------------------------+-------------------------------------------------------------------+
	| **beam**                      | True / False - generate an input gauss Twiss beam based on the    |
	|                               | values of the twiss parameters at the beginning of the lattice    |
	|                               | (startname) NOTE - we thoroughly recommend checking these         |
	|                               | parameters and this functionality is only for partial convenience |
	|                               | to have a model that works straight away.                         |
	+-------------------------------+-------------------------------------------------------------------+
	| **flipmagnets**               | True / False - flip the sign of all k values for magnets.         |
	|                               | Mad8 currently tracks particles agnostic of the particle charge.  |
	|                               | BDSIM however, follows the definition strictly :                  |
	|                               | Positive k implies horizontal focussing for positive particles    |
	|                               | therefore, vertical focussing for negative particles.             |
	|                               | Use this flag to flip the sign of all magnets.                    |
	+-------------------------------+-------------------------------------------------------------------+
	| **defaultAperSize**           | Aperture size to assu;e if none is specified. Default is 0.1m     |
	+-------------------------------+-------------------------------------------------------------------+
	| **defaultAperShape**          | Aperture shape to assume if none is specified.                    |
	|                               | Default is 'circular'.                                            |
	+-------------------------------+-------------------------------------------------------------------+
	| **biases**                    | Optional list of bias objects to be defined in own _bias.gmad     |
	|                               | file.  These can then be attached either with allelementdict for  |
	|                               | all components or userdict for individual ones.                   |
	+-------------------------------+-------------------------------------------------------------------+
	| **materials**                 | Optional list of material objects to be defined in own            |
	|                               | _materials.gmad file.  These can then be attached either with     |
	|                               | allelementdict for all components or userdict for individual ones.|
	+---------------------------------------------------------------------------------------------------+
	| **allelementdict**            | Dictionary of parameter/value pairs to be written to all elements |
	+-------------------------------+-------------------------------------------------------------------+
	| **optionsdict**               | Optional dictionary of general options to be written to the       |
	|                               | bdsim model options.                                              |
	+-------------------------------+-------------------------------------------------------------------+
	| **beamparamsdict**            | Optional dictionary of parameters to be passed to the beam.       |
	+-------------------------------+-------------------------------------------------------------------+
	| **linear**                    | Only linear optical components                                    |
	+-------------------------------+-------------------------------------------------------------------+
	| **overwrite**                 | Do not append an integer to the base file name if it already      |
	|                               | exists.  Instead overwrite the files.                             |
	+-------------------------------+-------------------------------------------------------------------+
	| **write**                     | Whether to write the converted machine to file or not.            |
	+-------------------------------+-------------------------------------------------------------------+
	| **namePrepend**               | Optional string prepended to the name of every component.         |
	+-------------------------------+-------------------------------------------------------------------+

	"""

	# test for inputfilename type
	if type(inputfilename) == str:
		twiss = _m8.Output(inputfilename)
		rmat = None
		if twiss.filetype != 'twiss':
			raise ValueError('Expect a twiss file to convert')
		data = twiss.data
	elif type(inputfilename) == dict:
		twiss = _m8.Output(inputfilename['twiss'])
		rmat = _m8.Output(inputfilename['rmat'], 'rmat')
		if twiss.filetype != 'twiss' or rmat.filetype != 'rmat':
			raise ValueError('Expect a twiss file and a rmat file to convert')
		data = twiss.data.merge(rmat.data)
	else:
		raise TypeError('Expect either a twiss file string or a dictionary with twiss and rmat file strings')

	# machine instance that will be added to
	machine = _Builder.Machine()

	# check if dictionaries are dictionaries
	varnames = ['collimatordict', 'userdict', 'partnamedict', 'allelementdict', 'optionsdict', 'beamparamsdict']
	vars     = [collimatordict,   userdict,  partnamedict,  allelementdict,  optionsdict,  beamparamsdict]
	for var, varname in zip(vars, varnames):
		typevar = type(var)
		if typevar not in (dict, _Data.BDSAsciiData):
			raise TypeError("Argument '" + varname + "' is not a dictionary")

	# can't use 'aperturedict' and 'aperlocalpositions' at the same time
	if aperturedict and aperlocalpositions:
		raise TypeError("'aperturedict' and 'aperlocalpositions' are mutually exclusive.")

	# try to check automatically if we need to flip magnets
	if "particletype" in beamparamsdict and flipmagnets is None:
		particletype = beamparamsdict['particletype']
		if particletype == "e-" or particletype == "electron":
			flipmagnets = True
			print('Detected electron in TFS file - changing flipmagnets to True')

	# If we have collimators but no collimator dict then inform that they will be converted to drifts
	if ("RCOL" in data['TYPE'] or "ECOL" in data['TYPE']) and not collimatordict:
		_warnings.warn("No collimatordict provided.  ALL collimators will be converted to DRIFTs.")

	if biases is not None:
		machine.AddBias(biases)

	if materials is not None:
		machine.AddMaterial(materials)

	# keep list of omitted zero length items
	itemsomitted = []

	# ==MAIN LOOP== #
	for item in data.iloc[startindex:endindex][::stepsize].iloc:
		index = item.name
		name = item['NAME']
		t = item['TYPE']
		l = item['L']
		
		# skip ignorable thin elements of zero length
		zerolength = True if l < 1e-9 else False
		if zerolength and t in _ignoreableThinElements:
			if verbose:
				print('skipping zero-length item: {}'.format(name))
			itemsomitted.append(name)
			continue

		gmadElement = _Mad82GmadElementFactory(item, allelementdict, verbose,userdict, collimatordict, partnamedict,
									flipmagnets, linear, zerolength, ignorezerolengthitems, namePrepend="", rmat=rmat)

		if gmadElement is None:  # factory returned nothing, go to next item.
			continue
		elif gmadElement.length == 0.0 and isinstance(gmadElement, _Builder.Drift):  # skip drifts of length 0
			continue
		elif l == 0 or name in collimatordict:  # Don't add apertures to thin elements or collimators
			machine.Append(gmadElement)
		elif aperlocalpositions and index in aperlocalpositions:  # split aperture if provided
			elements_split_with_aper = _GetElementSplitByAperture(gmadElement, aperlocalpositions[index])
			for ele in elements_split_with_aper:
				machine.Append(ele)
		else:  # Get element with single aperture
			element_with_aper = _GetSingleElementWithAper(twiss, item, gmadElement, aperturedict,
														  defaultAperSize, defaultAperShape)
			machine.Append(element_with_aper)

	if samplers is not None:  # Add Samplers
		machine.AddSampler(samplers)

	if beam:  # Add Beam
		bm = Mad82GmadBeam(data, startindex, verbose, extraParamsDict=beamparamsdict)
		machine.AddBeam(bm)

	options = _Options()  # Add Options
	if optionsdict:
		options.update(optionsdict)  # expand with user supplied bdsim options
	machine.AddOptions(options)

	if verbose:
		print('Total length: ', machine.GetIntegratedLength())
		print('Total angle:  ', machine.GetIntegratedAngle())
		print('items omitted: ')
		print(itemsomitted)
		print('number of omitted items: ', len(itemsomitted))

	if write:
		machine.Write(outputfilename, overwrite=overwrite)
	# We return machine twice to not break old interface of returning two machines.
	return machine, itemsomitted


def _Mad82GmadElementFactory(item, allelementdict, verbose,userdict, collimatordict, partnamedict,
							 flipmagnets, linear, zerolength, ignorezerolengthitems, namePrepend, rmat=None):
	"""
	Function which makes the correct GMAD element given a Mad8 element.
	"""
	# if it's already a prepared element, just return it
	if isinstance(item, _Builder.Element):
		return item

	factor = 1
	if flipmagnets is not None:
		factor = -1 if flipmagnets else 1  # flipping magnets

	kws = {}  # ensure empty
	kws = _deepcopy(allelementdict)  # deep copy as otherwise allelementdict gets irreparably changed!
	if verbose:
		print('Starting key word arguments from all element dict')
		print(kws)

	Type = item['TYPE']
	name = item['NAME']
	l = item['L']
	tilt = item['TILT']

	rname = pybdsim._General.PrepareReducedName(name)
	rname = namePrepend + rname
	
	# append any user defined parameters for this element into the kws dictionary
	if name in userdict:  # name appears in the madx.  try this first.
		kws.update(userdict[name])
	elif rname in userdict:  # rname appears in the gmad
		kws.update(userdict[rname])

	for partname in partnamedict:
		if partname in name:
			kws.update(partnamedict[partname])

	if verbose:
		print('Full set of key word arguments:')
		print(kws)

	if tilt != 0 and not _math.isnan(tilt):
		kws['tilt'] = tilt

	#######################################################################
	if Type == '    ':
		if not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'DRIF':
		return _Builder.Drift(rname, l, **kws)
	#######################################################################
	elif Type == 'MARK':
		if not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'SOLE':
		if l == 0.0 :
			return None # thin solenoid is omitted
		return _Builder.Solenoid(rname, l, ks=item['KS']/l, **kws)
	#######################################################################
	elif Type == 'INST':
		if zerolength and not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'MONI':
		if zerolength and not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'IMON':
		if not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'BLMO':
		if not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'WIRE':
		if not ignorezerolengthitems:
			return _Builder.Marker(rname)
	#######################################################################
	elif Type == 'QUAD':
		k1 = item['K1'] * factor
		return _Builder.Quadrupole(rname, l, k1, **kws)
	#######################################################################
	elif Type == 'SEXT':
		k2 = item['K2'] * factor if not linear else 0
		return _Builder.Sextupole(rname, l, k2, **kws)
	#######################################################################
	elif Type == 'OCTU':
		k3 = item['K3'] * factor if not linear else 0
		return _Builder.Octupole(rname, l, k3=k3, **kws)
	#######################################################################
	elif Type == 'DECU':
		pass
	#######################################################################
	elif Type == 'MULT':
		k1 = item['K1'] * factor
		k2 = item['K2'] * factor if not linear else 0
		k3 = item['K3'] * factor if not linear else 0
		ks = item['KS'] * factor
		if zerolength or l < _THIN_ELEMENT_THRESHOLD:
			return _Builder.ThinMultipole(rname, knl=(k1, k2, k3), ksl=ks, **kws)
		else:
			return _Builder.Multipole(rname, l, knl=(k1, k2, k3), ksl=ks, **kws)
	#######################################################################
	elif Type == 'HKIC':
		if verbose:
			print('HKICKER', rname)
		if not zerolength:
			if l > _THIN_ELEMENT_THRESHOLD:
				kws['l'] = l
		return _Builder.HKicker(rname, hkick=item['HKIC']*factor, **kws)
	#######################################################################
	elif Type == 'VKIC':
		if verbose:
			print('VKICKER', rname)
		if not zerolength:
			if l > _THIN_ELEMENT_THRESHOLD:
				kws['l'] = l
		return _Builder.VKicker(rname, vkick=item['VKIC']*factor, **kws)
	#######################################################################
	elif Type == 'KICK':
		if verbose:
			print('KICKER', rname)
		hkick = item['HKIC'] * factor
		vkick = item['VKIC'] * factor
		if not zerolength:
			if l > _THIN_ELEMENT_THRESHOLD:
				kws['l'] = l
		return _Builder.Kicker(rname, hkick=hkick, vkick=vkick, **kws)
	#######################################################################
	elif Type == 'SBEN':
		angle = item['ANGLE']
		e1 = item['E1'] if 'E1' in item else 0
		e2 = item['E2'] if 'E2' in item else 0
		h1 = item['H1'] if 'H1' in item else 0
		h2 = item['H2'] if 'H2' in item else 0
		k1 = item['K1']
		if e1 != 0:
			kws['e1'] = e1
		if e2 != 0:
			kws['e2'] = e2
		if h1 != 0:
			kws['h1'] = h1
		if h2 != 0:
			kws['h2'] = h2
		if k1 != 0:
			# NOTE we're not using factor for magnet flipping here
			kws['k1'] = k1
		return _Builder.SBend(rname, l, angle=angle, **kws)
	#######################################################################
	elif Type == 'RBEN':
		angle = item['ANGLE']
		e1 = item['E1'] if 'E1' in item else 0
		e2 = item['E2'] if 'E2' in item else 0
		h1 = item['H1'] if 'H1' in item else 0
		h2 = item['H2'] if 'H2' in item else 0
		k1 = item['K1']

		if angle != 0:  #protect against 0 angle rbends
			chordLength = 2 * (l / angle) * _np.sin(angle / 2.)
		else :
			# set element length to be the chord length - tfs output rbend length is arc length
			chordLength = l

		# subtract dipole angle/2 added on to poleface angles internally by madx
		poleInAngle = e1 - 0.5 * angle
		poleOutAngle = e2 - 0.5 * angle
		if poleInAngle != 0:
			kws['e1'] = poleInAngle
		if poleOutAngle != 0:
			kws['e2'] = poleOutAngle
		if h1 != 0:
			kws['h1'] = h1
		if h2 != 0:
			kws['h2'] = h2
		if k1 != 0:
			# NOTE we don't use factor here for magnet flipping
			kws['k1'] = k1
		return _Builder.RBend(rname, chordLength, angle=angle, **kws)
	#######################################################################
	elif Type == 'LCAV':
		volt = item['VOLT']
		freq = item['FREQ']
		lag = item['LAG']
		gradient = volt/l
		phase = lag * 2 * _np.pi

		if freq != 0:
			kws['freq'] = freq
		if phase != 0:
			kws['phase'] = phase
		return _Builder.RFCavity(rname, l, gradient=gradient, **kws)
	#######################################################################
	elif Type in {'ECOL', 'RCOL'}:
		if name in collimatordict:
			# gets a dictionary then extends kws dict with that dictionary
			colld = collimatordict[name]

			# collimator defined by external geometry file
			if 'geometryFile' in colld:
				kws['geometryFile'] = colld['geometryFile']
				k = 'outerDiameter'  # key
				if k not in kws:
					# not already specified via other dictionaries
					if k in colld:
						kws[k] = colld[k]
					else:
						kws[k] = 1  # ensure there's a default as not in madx
				return _Builder.ExternalGeometry(rname, l, **kws)
			else:
				kws['material'] = colld.get('material', 'copper')
				tilt = colld.get('tilt', 0)
				if tilt != 0:
					kws['tilt'] = tilt
				xsize = colld['xsize']
				ysize = colld['ysize']
				if verbose:
					print('collimator x,y size ', xsize, ysize)
				if 'outerDiameter' in colld:
					kws['outerDiameter'] = colld['outerDiameter']
				else:
					kws['outerDiameter'] = max([0.5, xsize * 2.5, ysize * 2.5])
				if Type == 'RCOL':
					return _Builder.RCol(rname, l, xsize, ysize, **kws)
				else:
					return _Builder.ECol(rname, l, xsize, ysize, **kws)
		# dict is incomplete or the component is erroneously
		# reffered to as a collimator even when it can be thought
		# of as a drift (e.g. LHC TAS).
		elif collimatordict != {}:
			msg = ("{} {} not found in collimatordict."
				" Will instead convert to a DRIFT!  This is not"
				" necessarily wrong!".format(Type, name))
			_warnings.warn(msg)
			return _Builder.Drift(rname, l, **kws)
		# if user didn't provide a collimatordict at all.
		else:
			return _Builder.Drift(rname, l, **kws)
	#######################################################################
	elif Type == 'MATR':
		if not isinstance('R11', item):
			raise ValueError('No Rmat file to extract element')
		index = item.name
		item = rmat.data.iloc[index]
		previtem = rmat.data.iloc[index-1]
		POSTMATRIX = _np.matrix([[item['R11'], item['R12'], item['R13'], item['R14']],
								 [item['R21'], item['R22'], item['R23'], item['R24']],
								 [item['R31'], item['R32'], item['R33'], item['R34']],
								 [item['R41'], item['R42'], item['R43'], item['R44']]])
		PRIORMATRIX = _np.matrix([[previtem['R11'], previtem['R12'], previtem['R13'], previtem['R14']],
								  [previtem['R21'], previtem['R22'], previtem['R23'], previtem['R24']],
								  [previtem['R31'], previtem['R32'], previtem['R33'], previtem['R34']],
								  [previtem['R41'], previtem['R42'], previtem['R43'], previtem['R44']]])

		RMAT = (POSTMATRIX*PRIORMATRIX.I)
		return _Builder.Rmat(rname, l, r11=RMAT[0, 0], r12=RMAT[0, 1], r13=RMAT[0, 2], r14=RMAT[0, 3],
							 		   r21=RMAT[1, 0], r22=RMAT[1, 1], r23=RMAT[1, 2], r24=RMAT[1, 3],
							 		   r31=RMAT[2, 0], r32=RMAT[2, 1], r33=RMAT[2, 2], r34=RMAT[2, 3],
							 		   r41=RMAT[3, 0], r42=RMAT[3, 1], r43=RMAT[3, 2], r44=RMAT[3, 3], **kws)
	#######################################################################
	else:
		print('unknown element type:', Type, 'for element named: ', name)
		if zerolength and not ignorezerolengthitems:
			print('putting marker in instead as its zero length')
			return _Builder.Marker(rname)
		print('putting drift in instead as it has a finite length')
		return _Builder.Drift(rname, l)
	#######################################################################


def _GetElementSplitByAperture(gmadElement, localApertures):
	"""Return an element with splited aperture"""
	# tolerate any bad apertures - like only a few - and just don't append them
	apertures = []
	for point, aper in localApertures:
		try:
			arp = _Builder.PrepareApertureModel(aper, warningName=gmadElement.name)
			apertures.append(arp)
		except ValueError:
			pass

	if localApertures[0][0] != 0.0:
		raise ValueError("No aperture defined at start of element.")
	if len(localApertures) > 1:
		split_points = [point for point, _ in localApertures[1:]]
		split_elements = gmadElement.split(split_points)
		for aper, split_element in zip(apertures, split_elements):
			split_element.update(aper)
		return split_elements
	elif len(localApertures) == 1:
		gmadElement = _deepcopy(gmadElement)
		gmadElement.update(apertures[0])
		return [gmadElement]
	raise ValueError("Unable to split element by apertures.")


def _GetSingleElementWithAper(twiss, item, gmadElement, aperturedict, defaultAperSize, defaultAperShape):
	"""Return an element with unsplit aperture"""
	gmadElement = _deepcopy(gmadElement)
	name = item["NAME"]
	index = item.name
	aper = {}
	if name in aperturedict:
		aper = _Builder.PrepareApertureModel(aperturedict[name], defaultAperShape)
	else:
		this_aperdict = {'APER_1' : twiss.getAperture(index, defaultAperSize)}
		aper = _Builder.PrepareApertureModel(this_aperdict, defaultAperShape)

	gmadElement.update(aper)
	return gmadElement


def Mad82GmadBeam(data, startindex=0, verbose=False, extraParamsDict={}):
	"""
	Takes a pymad8 dataframe and extracts beam information from extraParamsDict
	to create a BDSIM beam definition in a pybdsim.Beam object.
	Note that if kwarg startname is used, the optics are retrieved at the
	start of the element, i.e. you do not need to get the optics of
	the previous element, this function does that automatically.

	Works for e+, e- and proton.

	"""

	# MADX defines parameters at the end of elements so need to go 1 element back if we can
	if startindex > 0:
		startindex -= 1

	energy = data['E'][startindex]
	if 'EX' not in extraParamsDict or 'EY' not in extraParamsDict:
		raise ValueError('Missing emittance description in extraParamsDict')
	if 'Esprd' not in extraParamsDict:
		raise ValueError('Missing energy spread in extraParamsDict')
	if 'particletype' not in extraParamsDict:
		raise ValueError('Missing particle type in extraParamsDict')

	ex =		extraParamsDict.pop('EX')
	ey =		extraParamsDict.pop('EY')
	sigmae =	extraParamsDict.pop('Esprd')
	particle =	extraParamsDict.pop('particletype')

	if particle != 'e-' and particle != 'e+' and particle != 'proton':
		raise ValueError("Unsupported particle : " + particle)
		
	# return _Beam.Beam('e-',16.5,'gausstwiss')
	beam = _Beam.Beam(particle, energy, 'gausstwiss')
	if ex != 0:
		beam._SetEmittanceX(ex, 'm')
	if ey != 0:
		beam._SetEmittanceY(ey, 'm')
	beam._SetSigmaE(sigmae)

	beamparams = {"SetBetaX": 'BETX', "SetBetaY": 'BETY', "SetAlphaX": 'ALPHX', "SetAlphaY": 'ALPHY',
				  "SetDispX":'DX', "SetDispY": 'DY', "SetDispXP": 'DPX', "SetDispYP": 'DPY',
				  "SetXP0": 'PX', "SetYP0": 'PY', "SetX0": 'X', "SetY0": 'Y'}

	for func, parameter in beamparams.items():
		if parameter in list(data.keys()):
			getattr(beam, func)(data[parameter][startindex])

	for k, v in extraParamsDict.items():
		beam[k] = v

	return beam
